import os,sys,json,glob,shutil,uuid,time
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'pbs'
from .context import LocalSession
from .context import LocalContext
from .context import PBS
from .context import JobStatus
from .context import Dispatcher
from .context import my_file_cmp
from .context import setUpModule

@unittest.skipIf(not shutil.which("qsub"), "requires PBS")
class TestDispatcher(unittest.TestCase) :
    def setUp(self) :
        os.makedirs('loc', exist_ok = True)
        os.makedirs('rmt', exist_ok = True)
        os.makedirs('loc/task0', exist_ok = True)
        os.makedirs('loc/task1', exist_ok = True)
        os.makedirs('loc/task2', exist_ok = True)
        for ii in ['loc/task0', 'loc/task1', 'loc/task2']:
            with open(os.path.join(ii, 'test0'),'w') as fp:
                fp.write('this is test0 from ' + ii + '\n')
        work_profile = {'work_path':'rmt'}
        self.disp = Dispatcher(work_profile, 'local', 'pbs')

    def test_sub_success(self):
        tasks = ['task0', 'task1', 'task2']
        self.disp.run_jobs(None,
                           'cp test0 test1',
                           'loc',
                           tasks,
                           2,
                           [],
                           ['test0'],
                           ['test1', 'hereout.log', 'hereerr.log'],
                           outlog = 'hereout.log',
                           errlog = 'hereerr.log')
        for ii in tasks:            
            my_file_cmp(self,
                        os.path.join('loc', ii, 'test0'),
                        os.path.join('loc', ii, 'test1'))
            self.assertTrue(os.path.isfile(os.path.join('loc', ii, 'hereout.log')))
            self.assertTrue(os.path.isfile(os.path.join('loc', ii, 'hereerr.log')))
            



import sys, os, hashlib, pathlib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..' )))
# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))
import dpdispatcher
# from dpdispatcher.local_context import LocalSession
from dpdispatcher.local_context import LocalContext
# from dpdispatcher.local_context import local_context
from dpdispatcher.lazy_local_context import LazyLocalContext
from dpdispatcher.ssh_context import SSHSession
from dpdispatcher.ssh_context import SSHContext
# from dpgen.dispatcher.Dispatcher import FinRecord
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.pbs import PBS
from dpdispatcher.slurm import Slurm
from dpdispatcher.shell import Shell
from dpdispatcher.lsf import LSF
from dpdispatcher.base_context import BaseContext
from dpdispatcher.distributed_shell import DistributedShell
from dpdispatcher.hdfs_context import HDFSContext
from dpdispatcher.hdfs_cli import HDFS

from dpdispatcher.local_context import _identical_files
try:
    from dpdispatcher.dp_cloud_server import DpCloudServer
    from dpdispatcher.dp_cloud_server_context import DpCloudServerContext
except:
    pass
from dpdispatcher.submission import Submission, Job, Task, Resources
from dpdispatcher.machine import Machine
import dargs

def setUpModule():
    os.chdir(os.path.abspath(os.path.dirname(__file__)))

def get_file_md5(file_path):
    return hashlib.md5(pathlib.Path(file_path).read_bytes()).hexdigest()



import os,sys,json,glob,shutil,uuid,time
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
# from .context import LocalSession
# from .context import LocalContext
from .context import LocalContext
from .context import PBS
from .context import JobStatus
from .context import setUpModule
from .context import Submission, Job, Task, Resources
from .sample_class import SampleClass

class TestResources(unittest.TestCase) :
    def setUp(self) :
        self.maxDiff = None
        self.resources = SampleClass.get_sample_resources()
        self.resources_dict = SampleClass.get_sample_resources_dict()

    def test_eq(self):
        self.assertEqual(self.resources, SampleClass.get_sample_resources())

    def test_serialize(self):
        self.assertEqual(self.resources.serialize(), self.resources_dict)

    def test_deserialize(self):
        resources = Resources.deserialize(resources_dict=self.resources_dict)
        self.assertEqual(self.resources, resources)
    
    def test_serialize_deserialize(self):
        self.assertEqual(self.resources, Resources.deserialize(resources_dict=self.resources.serialize()))
        
    def test_resources_json(self):
        with open('jsons/resources.json', 'r') as f:
            resources_json_dict = json.load(f)
        self.assertTrue(resources_json_dict, self.resources_dict)
        self.assertTrue(resources_json_dict, self.resources.serialize())

    def test_load_from_json(self):
        resources = Resources.load_from_json('jsons/resources.json')
        self.assertTrue(resources, self.resources)





import os,sys,json,glob,shutil,uuid,time
import unittest
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import LocalContext, LocalSession
from .context import setUpModule
from .context import _identical_files

class TestIdFile(unittest.TestCase) :
    def test_id(self) :
        with open('f0', 'w') as fp:
            fp.write('foo')
        with open('f1', 'w') as fp:
            fp.write('foo')
        self.assertTrue(_identical_files('f0', 'f1'))
        os.remove('f0')
        os.remove('f1')

    def test_diff(self) :
        with open('f0', 'w') as fp:
            fp.write('foo')
        with open('f1', 'w') as fp:
            fp.write('bar')
        self.assertFalse(_identical_files('f0', 'f1'))
        os.remove('f0')
        os.remove('f1')


class TestLocalContext(unittest.TestCase):
    def setUp(self) :
        os.makedirs('loc', exist_ok = True)
        os.makedirs('loc/task0', exist_ok = True)
        os.makedirs('loc/task1', exist_ok = True)
        for ii in ['loc/task0', 'loc/task1']:
            with open(os.path.join(ii, 'test0'),'w') as fp:
                fp.write(str(uuid.uuid4()))
            with open(os.path.join(ii, 'test1'),'w') as fp:
                fp.write(str(uuid.uuid4()))
            with open(os.path.join(ii, 'test2'),'w') as fp:
                fp.write(str(uuid.uuid4()))
            os.makedirs(os.path.join(ii, 'dir0'), exist_ok = True)
            os.makedirs(os.path.join(ii, 'dir2'), exist_ok = True)
            with open(os.path.join(ii, 'dir2', 'dtest0'),'w') as fp:
                fp.write(str(uuid.uuid4()))
        os.makedirs('rmt', exist_ok = True)

    def tearDown(self):
        shutil.rmtree('loc')
        shutil.rmtree('rmt')

    def test_upload_non_exist(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        # test uploading non-existing file
        with self.assertRaises(RuntimeError):
            self.job.upload(tasks, ['foo'])

    def test_upload(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        self.job1 = LocalContext('loc', work_profile, job_uuid = self.job.job_uuid)
        tasks = ['task0', 'task1']
        files = ['test0', 'test1', 'dir2/dtest0']
        self.job.upload(tasks, files)
        for ii in tasks :
            for jj in files :
                locf = os.path.join('loc', ii, jj)
                rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj)
                with open(locf) as fp:
                    locs = fp.read()
                with open(rmtf) as fp:
                    rmts = fp.read()
                self.assertEqual(locs, rmts)
        self.job.upload(tasks, ['dir0'])
        for ii in tasks :
            for jj in ['dir0'] :
                locf = os.path.join('loc', ii, jj)
                rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj)
                self.assertEqual(os.path.realpath(locf),
                                 os.path.realpath(rmtf))
        self.job1.upload(tasks, files)
        for ii in tasks :
            for jj in files :
                locf = os.path.join('loc', ii, jj)
                rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj)
                with open(locf) as fp:
                    locs = fp.read()
                with open(rmtf) as fp:
                    rmts = fp.read()
                self.assertEqual(locs, rmts)

    def test_dl_f_f(self):
        # no local, no remote
        self.test_download_non_exist()
        
    def test_dl_t_f(self) :
        # has local, no remote
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        record_uuid = []        
        for ii in tasks :
            for jj in ['dir1'] :
                os.makedirs(os.path.join('loc',ii,jj), exist_ok=False)
                for kk in ['test6', 'test7']:
                    with open(os.path.join('loc',ii,jj,kk), 'w') as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
                        record_uuid.append(tmp)
        files = ['dir1']
        self.job.download(tasks, files)
        cc = 0
        for ii in tasks :
            for jj in ['dir1'] :
                for kk in ['test6', 'test7']:
                    with open(os.path.join('loc',ii,jj,kk), 'r') as fp:
                        tmp = fp.read()
                        self.assertEqual(tmp, record_uuid[cc])
                        cc += 1
        
    def test_dl_t_t(self) :
        # has local, has remote
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        for ii in tasks :
            for jj in ['dir1'] :
                os.makedirs(os.path.join('loc',ii,jj), exist_ok=False)
                for kk in ['test6', 'test7']:
                    with open(os.path.join('loc',ii,jj,kk), 'w') as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
        record_uuid = []        
        for ii in tasks :
            for jj in ['dir1'] :
                os.makedirs(os.path.join('rmt', self.job.job_uuid,ii,jj), exist_ok=False)
                for kk in ['test6', 'test7']:
                    with open(os.path.join('rmt', self.job.job_uuid,ii,jj,kk), 'w') as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
                        record_uuid.append(tmp)
        files = ['dir1']
        self.job.download(tasks, files)
        cc = 0
        for ii in tasks :
            for jj in ['dir1'] :
                for kk in ['test6', 'test7']:
                    with open(os.path.join('loc',ii,jj,kk), 'r') as fp:
                        tmp = fp.read()
                        self.assertEqual(tmp, record_uuid[cc])
                        cc += 1
        

    def test_download_non_exist(self):
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        # down load non-existing file
        with self.assertRaises(RuntimeError):
            self.job.download(tasks, ['foo'])        

    def test_download(self):        
        # upload files
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test4', 'test5'] :
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # generate extra donwload dirs and files
        for ii in tasks :
            for jj in ['dir1'] :
                os.makedirs(os.path.join('rmt',self.job.job_uuid,ii,jj), exist_ok=False)
                for kk in ['test6']:
                    with open(os.path.join('rmt',self.job.job_uuid,ii,jj,kk), 'w') as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
                        record_uuid.append(tmp)                        
        # donwload
        files = ['test0', 'dir0', 'test4', 'test5', 'dir1']
        self.job.download(tasks, files)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test4', 'test5'] :
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                for kk in ['test6']:
                    with open(os.path.join('loc',ii,jj,kk), 'r') as fp:
                        tmp = fp.read()
                        self.assertEqual(tmp, record_uuid[cc])
                        cc += 1
        # check links preserved
        for ii in tasks :
            for jj in ['test0'] :
                locf = os.path.join('loc', ii, jj)
                rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj)
                self.assertEqual(os.path.realpath(locf),
                                 os.path.realpath(rmtf))
        for ii in tasks :
            for jj in ['dir0'] :
                for kk in ['test6'] :
                    locf = os.path.join('loc', ii, jj, kk)
                    rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj, kk)
                    self.assertEqual(os.path.realpath(locf),
                                     os.path.realpath(rmtf))

    def test_download_check_mark(self):        
        # upload files
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test7', 'test8'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6') :
                    continue
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ['test7', 'test8', 'dir1']
        self.job.download(tasks, files, check_exists = True, mark_failure = True)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test7', 'test8'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6') :
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, jj))
                    self.assertTrue(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)), 
                                    msg = 'failed to find ' + os.path.join('loc', ii, 'tag_failure_download_%s' % jj))
                    continue
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)))
                self.assertTrue(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)))


    def test_download_check_nomark(self):        
        # upload files
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test7', 'test8'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6') :
                    continue
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ['test7', 'test8', 'dir1']
        self.job.download(tasks, files, check_exists = True, mark_failure = False)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test7', 'test8'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6') :
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, jj))
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, 'tag_failure_download_%s' % jj))
                    continue
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)))
                self.assertFalse(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)))

                
    def test_block_call(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        files = ['test0', 'test1']
        self.job.upload(tasks, files)
        # ls
        code, stdin, stdout, stderr = self.job.block_call('ls')
        self.assertEqual(stdout.read().decode('utf-8'), 'task0\ntask1\n')
        self.assertEqual(stdout.readlines(), ['task0\n','task1\n'])
        self.assertEqual(code, 0)
        code, stdin, stdout, stderr = self.job.block_call('ls a')
        self.assertEqual(code, 2)
        # self.assertEqual(stderr.read().decode('utf-8'), 'ls: cannot access a: No such file or directory\n')
        err_msg = stderr.read().decode('utf-8')
        self.assertTrue('ls: cannot access' in err_msg)
        self.assertTrue('No such file or directory\n' in err_msg)


    def test_block_checkcall(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job  = LocalContext('loc', work_profile)
        tasks = ['task0', 'task1']
        files = ['test0', 'test1']
        self.job.upload(tasks, files)
        # ls
        stdin, stdout, stderr = self.job.block_checkcall('ls')
        self.assertEqual(stdout.read().decode('utf-8'), 'task0\ntask1\n')
        self.assertEqual(stdout.readlines(), ['task0\n','task1\n'])
        with self.assertRaises(RuntimeError):
            stdin, stdout, stderr = self.job.block_checkcall('ls a')
            
    def test_file(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job = LocalContext('loc', work_profile)
        self.assertFalse(self.job.check_file_exists('aaa'))
        tmp = str(uuid.uuid4())
        self.job.write_file('aaa', tmp)
        self.assertTrue(self.job.check_file_exists('aaa'))
        tmp1 = self.job.read_file('aaa')
        self.assertEqual(tmp, tmp1)
        

    def test_call(self) :
        work_profile = LocalSession({'work_path':'rmt'})
        self.job = LocalContext('loc', work_profile)
        proc = self.job.call('sleep 3')
        self.assertFalse(self.job.check_finish(proc))
        time.sleep(1)
        self.assertFalse(self.job.check_finish(proc))
        time.sleep(2.5)
        self.assertTrue(self.job.check_finish(proc))
        r,o,e=self.job.get_return(proc)
        self.assertEqual(r, 0)
        self.assertEqual(o.read(), b'')
        self.assertEqual(e.read(), b'')
        r,o,e=self.job.get_return(proc)
        self.assertEqual(r, 0)
        self.assertEqual(o, None)
        self.assertEqual(e, None)




#import psutil

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.machine import Machine

shell_script_header_template="""
#!/bin/bash -l
"""

class Shell(Machine):
    def gen_script(self, job):
        shell_script = super(Shell, self).gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        shell_script_header = shell_script_header_template
        return shell_script_header

    def do_submit(self, job):
        script_str = self.gen_script(job) 
        script_file_name = job.script_file_name
        job_id_name = job.job_hash + '_job_id'
        output_name = job.job_hash + '.out'
        self.context.write_file(fname=script_file_name, write_str=script_str)
        ret, stdin, stdout, stderr = self.context.block_call('cd %s && { nohup bash %s 1>>%s 2>>%s & } && echo $!' % (self.context.remote_root, script_file_name, output_name, output_name) )
        if ret != 0:
            err_str = stderr.read().decode('utf-8')
            raise RuntimeError\
                    ("status command squeue fails to execute\nerror message:%s\nreturn code %d\n" % (err_str, ret))
        job_id = int(stdout.read().decode('utf-8').strip())
        self.context.write_file(job_id_name, str(job_id))
        return job_id

        # script_file_name = job.script_file_name
        # script_str = self.gen_script(job)
        # job_id_name = job.job_hash + '_job_id'
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        # self.context.write_file(fname=script_file_name, write_str=script_str)
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        # subret = (stdout.readlines())
        # job_id = subret[0].split()[0]
        # self.context.write_file(job_id_name, job_id)        
        # return job_id


    def default_resources(self, resources) :
        pass
    
    def check_status(self, job):
        job_id = job.job_id
        # print('shell.check_status.job_id', job_id)
        # job_state = JobStatus.unknown
        if job_id == "" :
            return JobStatus.unsubmitted

        ret, stdin, stdout, stderr = self.context.block_call(f"if ps -p {job_id} > /dev/null; then echo 1; fi")
        if ret != 0:
            err_str = stderr.read().decode('utf-8')
            raise RuntimeError\
                    ("status command squeue fails to execute\nerror message:%s\nreturn code %d\n" % (err_str, ret))
        
        if_job_exists = bool(stdout.read().decode('utf-8').strip())
        if self.check_finish_tag(job=job):
            dlog.info(f"job: {job.job_hash} {job.job_id} finished")
            return JobStatus.finished

        if if_job_exists:
            return JobStatus.running
        else:
            return JobStatus.terminated
        # return job_state
    
    # def check_status(self, job):
    #     job_id = job.job_id
    #     uuid_names = job.job_hash
    #     cnt = 0
    #     ret, stdin, stdout, stderr = self.context.block_call("ps aux | grep %s"%uuid_names)
    #     response_list = stdout.read().decode('utf-8').split("\n")
    #     for response in response_list:
    #         if  uuid_names + ".sub" in response:
    #             return True
    #     return False

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        # print('job finished: ',job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)



from enum import IntEnum
class JobStatus(IntEnum) :
    unsubmitted = 1
    waiting = 2
    running = 3
    terminated = 4
    finished = 5
    completing = 6
    unknown = 100
#     def __str__(self):
#         return repr(self)




# from dpdispatcher.batch_object import BatchObject
# from dpdispatcher.batch import Batch
import sys, os, shutil
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import Submission, Job, Task, Resources
from .context import Shell
from .context import LocalContext
from .context import get_file_md5
from .context import Machine

import unittest
import json


class TestShellTrival(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        # self.local_context_dict = {
        #     'local_root': './test_shell_trival_dir',
        #     'remote_root': './tmp_shell_trival_dir'
        # }
    
    def test_shell_trival(self):
        with open('jsons/machine_local_shell.json', 'r') as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict['machine'])
        resources = Resources(**machine_dict['resources'])

        task1 = Task(command='cat example.txt', task_work_path='dir1/', forward_files=['example.txt'], backward_files=['out.txt'], outlog='out.txt')
        task2 = Task(command='cat example.txt', task_work_path='dir2/', forward_files=['example.txt'], backward_files=['out.txt'], outlog='out.txt')
        task3 = Task(command='cat example.txt', task_work_path='dir3/', forward_files=['example.txt'], backward_files=['out.txt'], outlog='out.txt')
        task4 = Task(command='cat example.txt', task_work_path='dir4/', forward_files=['example.txt'], backward_files=['out.txt'], outlog='out.txt')
        task_list = [task1, task2, task3, task4]

        submission = Submission(work_base='parent_dir/',
            machine=machine,
            resources=resources,
            forward_common_files=['graph.pb'],
            backward_common_files=[],
            task_list=task_list
        )
        submission.run_submission(clean=False)

        for dir in ['dir1', 'dir2', 'dir3', 'dir4']:
            f1 = os.path.join('test_shell_trival_dir/', 'parent_dir/', dir, 'example.txt')
            f2 = os.path.join('test_shell_trival_dir/', 'parent_dir/', dir, 'out.txt')
            self.assertEqual(get_file_md5(f1), get_file_md5(f2))

    def test_shell_fail(self):
        with open('jsons/machine_local_shell.json', 'r') as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict['machine'])
        resources = Resources(**machine_dict['resources'])

        task = Task(command='cat mock_fail_task.txt && exit 1',
            task_work_path='./',
            forward_files=[], 
            backward_files=['out.txt'],
            outlog='out.txt')

        task_list = [task,]

        submission = Submission(work_base='fail_dir/',
            machine=machine,
            resources=resources,
            forward_common_files=[],
            backward_common_files=[],
            task_list=task_list
        )
        with self.assertRaises(RuntimeError):
            submission.run_submission()

    def test_shell_recover(self):
        with open('jsons/machine_lazylocal_shell.json', 'r') as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict['machine'])
        resources = Resources(**machine_dict['resources'])

        pass


    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_shell_trival_dir/')








import sys, os, json
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..' )))
# from .context import dpdispatcher
#from dpdispatcher.local_context import LocalContext
from dpdispatcher.lazy_local_context import LazyLocalContext
# from dpdispatcher.ssh_context import SSHContext

from dpdispatcher.submission import Submission, Job, Task, Resources
from dpdispatcher.machine import Machine
# from dpdispatcher.submission import
# from dpdispatcher.slurm import Slurm

# local_session = LocalSession({'work_path':'temp2'})
# local_context = LocalContext(local_root='test_slurm_dir/', work_profile=local_session)
# lazy_local_context = LazyLocalContext(local_root='test_slurm_dir/')


# machine_dict = dict(hostname='localhost', remote_root='/home/dp/dpdispatcher/tests/temp2', username='dp')
# ssh_session = SSHSession(**machine_dict)
# ssh_session = SSHSession(hostname='8.131.233.55', remote_root='/home/dp/dp_remote', username='dp')
# ssh_context = SSHContext(local_root='test_slurm_dir', ssh_session=ssh_session)
# slurm = Slurm(context=ssh_context)
# slurm = Slurm(context=lazy_local_context)

# resources = Resources(number_node=1, cpu_per_node=4, gpu_per_node=2, queue_name="GPU_2080Ti", group_size=4, 
#     custom_flags=['#SBATCH --exclude=2080ti000,2080ti001,2080ti002,2080ti004,2080ti005,2080ti006'],
#     para_deg=2,
#     strategy={"if_cuda_multi_devices":True})
# slurm_sbatch_dict={'mem': '10G', 'cpus_per_task':1, 'time': "120:0:0"} 
# slurm_resources = SlurmResources(resources=resources, slurm_sbatch_dict=slurm_sbatch_dict)

with open("jsons/machine_slurm.json", 'r') as f:
    mdata = json.load(f)

machine = Machine.load_from_dict(mdata['machine'])
resources = Resources.load_from_dict(mdata['resources'])

submission = Submission(work_base='0_md/', machine=machine, resources=resources,  forward_common_files=['graph.pb'], backward_common_files=[]) #,  batch=PBS)
task1 = Task(command='lmp -i input.lammps', task_work_path='bct-1/', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps']) 
task2 = Task(command='lmp -i input.lammps', task_work_path='bct-2/', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'])
task3 = Task(command='lmp -i input.lammps', task_work_path='bct-3/', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'])
task4 = Task(command='lmp -i input.lammps', task_work_path='bct-4/', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'])
submission.register_task_list([task1, task2, task3, task4, ])
submission.run_submission(clean=True)



import os,sys,json,glob,shutil,uuid,getpass
import unittest
import pathlib
from paramiko.ssh_exception import NoValidConnectionsError
from paramiko.ssh_exception import SSHException

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import SSHContext, SSHSession
from .context import Machine
from .sample_class import SampleClass

class TestSSHContext(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with open('jsons/machine_ali_ehpc.json', 'r') as f:
            mdata = json.load(f)
        try:
            cls.machine = Machine.load_from_dict(mdata['machine'])
        except SSHException:
            raise unittest.SkipTest("SSHException ssh cannot connect")
        cls.submission = SampleClass.get_sample_submission()
        cls.submission.bind_machine(cls.machine)
        cls.submission_hash = cls.submission.submission_hash
        file_list = ['bct-1/log.lammps', 'bct-2/log.lammps', 'bct-3/log.lammps', 'bct-4/log.lammps']
        for file in file_list:
            cls.machine.context.write_file(file, '# mock log')
    
    def setUp(self):
        self.context = self.__class__.machine.context

    def test_ssh_session(self):
        self.assertIsInstance(
            self.__class__.machine.context.ssh_session, SSHSession
        )

    def test_upload(self):
        self.context.upload(self.__class__.submission)
        check_file_list = ['graph.pb', 'bct-1/conf.lmp', 'bct-4/input.lammps']
        for file in check_file_list:
            self.assertTrue(self.context.check_file_exists(os.path.join(self.context.remote_root, file)))

    def test_download(self):
        self.context.download(self.__class__.submission)
        





import hashlib
import time
import struct
import hmac
import base64
import subprocess

def get_sha256(filename):
    """Get sha256 of a file.
    
    Parameters
    ----------
    filename: str
        The filename.

    Returns
    -------
    sha256: str
        The sha256.
    """
    h = hashlib.sha256()
    # buffer size: 128 kB
    b = bytearray(128*1024)
    mv = memoryview(b)
    with open(filename, 'rb', buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    sha256 = h.hexdigest()
    return sha256


def generate_totp(secret: str, period: int=30, token_length: int=6) -> int:
    """Generate time-based one time password (TOTP) from the secret.

    Some HPCs use TOTP for two-factor authentication for safety.

    Parameters
    ----------
    secret: str
        The encoded secret provided by the HPC. It's usually extracted
        from a 2D code and base32 encoded.
    period: int, default=30
        Time period where the code is valid in seconds.
    token_length: int, default=6
        The token length.
    
    Returns
    -------
    token: int
        The generated token.

    References
    ----------
    https://github.com/lepture/otpauth/blob/49914d83d36dbcd33c9e26f65002b21ce09a6303/otpauth.py#L143-L160
    """
    timestamp = time.time()
    counter = int(timestamp) // period
    msg = struct.pack('>Q', counter)
    digest = hmac.new(base64.b32decode(secret), msg, hashlib.sha1).digest()
    ob = digest[19]
    pos = ob & 15
    base = struct.unpack('>I', digest[pos:pos + 4])[0] & 0x7fffffff
    token = base % (10**token_length)
    return str(token).zfill(token_length)

def run_cmd_with_all_output(cmd):
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    ret = proc.returncode
    return (ret, out, err)


import shutil

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.dpcloudserver import zip_file
from dpdispatcher.machine import Machine
from dpdispatcher.dpcloudserver.api import API
from dpdispatcher.dpcloudserver.config import ALI_OSS_BUCKET_URL
import time
import warnings
import os
import uuid

shell_script_header_template="""
#!/bin/bash -l
"""

class DpCloudServer(Machine):
    def __init__(self, context):
        self.context = context
        self.input_data = context.remote_profile['input_data'].copy()
        self.api_version = 2
        if 'api_version' in self.input_data:
            self.api_version = self.input_data.get('api_version', 2)
        if 'lebesgue_version' in self.input_data:
            self.api_version = self.input_data.get('lebesgue_version', 2)
        self.grouped = self.input_data.get('grouped', False)
        email = context.remote_profile.get("email", None)
        username = context.remote_profile.get('username', None)
        password = context.remote_profile.get('password', None)
        if email is None and username is not None:
            raise DeprecationWarning("username is no longer support in current version, "
                                     "please consider use email instead of username.")
        if email is None:
            raise ValueError("can not find email in remote_profile, please check your machine file.")
        if password is None:
            raise ValueError("can not find password in remote_profile, please check your machine file.")
        if self.api_version == 1:
            raise DeprecationWarning('api version 1 is deprecated. Use version 2 instead.')
        self.api = API(email, password)
        self.group_id = None

    def gen_script(self, job):
        shell_script = super(DpCloudServer, self).gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        shell_script_header = shell_script_header_template
        return shell_script_header

    def gen_local_script(self, job):
        script_str = self.gen_script(job)
        script_file_name = job.script_file_name
        self.context.write_local_file(
            fname=script_file_name,
            write_str=script_str
        )
        return script_file_name

    def _gen_backward_files_list(self, job):
        result_file_list = []
        # result_file_list.extend(job.backward_common_files)
        for task in job.job_task_list:
            result_file_list.extend([ os.path.join(task.task_work_path,b_f) for b_f in task.backward_files])
        return result_file_list

    def _gen_oss_path(self, job, zip_filename):
        if hasattr(job, 'upload_path') and job.upload_path:
            return job.upload_path
        else:
            program_id = self.context.remote_profile.get('program_id')
            if program_id is None:
                dlog.info("can not find program id in remote profile, upload to default program id.")
                program_id = 0
            uid = uuid.uuid4()
            path = os.path.join("program", str(program_id), str(uid), zip_filename)
            setattr(job, 'upload_path', path)
            return path

    def do_submit(self, job):
        self.gen_local_script(job)
        zip_filename = job.job_hash + '.zip'
        # oss_task_zip = 'indicate/' + job.job_hash + '/' + zip_filename
        oss_task_zip = self._gen_oss_path(job, zip_filename)
        job_resources = ALI_OSS_BUCKET_URL + oss_task_zip

        input_data = self.input_data.copy()

        input_data['job_resources'] = job_resources
        input_data['command'] = f"bash {job.script_file_name}"
        # input_data['backward_files'] = self._gen_backward_files_list(job)
        if self.context.remote_profile.get('program_id') is None:
            warnings.warn('program_id will be compulsory in the future.')
        job_id, group_id = self.api.job_create(
            job_type=input_data['job_type'],
            oss_path=input_data['job_resources'],
            input_data=input_data,
            program_id=self.context.remote_profile.get('program_id', None),
            group_id=self.group_id
        )
        if self.grouped:
            self.group_id = group_id
        job.job_id = str(job_id) + ':job_group_id:' + str(group_id)
        job_id = job.job_id
        job.job_state = JobStatus.waiting
        return job_id

    def check_status(self, job):
        if job.job_id == '':
            return JobStatus.unsubmitted
        job_id = job.job_id
        group_id = None
        if isinstance(job.job_id, str) and ':job_group_id:' in job.job_id:
            group_id = None
            ids = job.job_id.split(":job_group_id:")
            job_id, group_id = int(ids[0]), int(ids[1])
            if self.input_data.get('grouped') and ':job_group_id:' not in self.input_data:
                self.group_id = group_id
            self.api_version = 2
        dlog.debug(f"debug: check_status; job.job_id:{job_id}; job.job_hash:{job.job_hash}")
        check_return = None
        # print("api",self.api_version,self.input_data.get('job_group_id'),job.job_id)
        check_return = self.api.get_tasks(job_id,group_id)
        try:
            dp_job_status = check_return["status"]
        except IndexError as e:
            dlog.error(f"cannot find job information in check_return. job {job.job_id}. check_return:{check_return}; retry one more time after 60 seconds")
            time.sleep(60)
            retry_return = self.api.get_tasks(job_id, group_id)
            try:
                dp_job_status = retry_return["status"]
            except IndexError as e:
                raise RuntimeError(f"cannot find job information in dpcloudserver's database for job {job.job_id} {check_return} {retry_return}")

        job_state = self.map_dp_job_state(dp_job_status)
        if job_state == JobStatus.finished:
            self._download_job(job)
        return job_state


    def _download_job(self, job):
        job_url = self.api.get_job_result_url(job.job_id)
        if not job_url:
            return
        job_hash = job.job_hash
        result_filename = job_hash + '_back.zip'
        target_result_zip = os.path.join(self.context.local_root, result_filename)
        self.api.download_from_url(job_url, target_result_zip)
        zip_file.unzip_file(target_result_zip, out_dir=self.context.local_root)
        try:
            os.makedirs(os.path.join(self.context.local_root, 'backup'), exist_ok=True)
            shutil.move(target_result_zip,
                        os.path.join(self.context.local_root, 'backup', os.path.split(target_result_zip)[1]))
        except (OSError, shutil.Error) as e:
            dlog.exception("unable to backup file, " + str(e))

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        dlog.info('check if job finished: ',job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)
        # return
        # pass

    def check_if_recover(self, submission):
        return False
        # pass

    @staticmethod
    def map_dp_job_state(status):
        if isinstance(status, JobStatus):
            return status
        map_dict = {
            -1:JobStatus.terminated,
            0:JobStatus.waiting,
            1:JobStatus.running,
            2:JobStatus.finished,
            3:JobStatus.waiting,
            4:JobStatus.running,
            5:JobStatus.terminated
        }
        if status not in map_dict:
            dlog.error(f'unknown job status {status}')
            return JobStatus.unknown
        return map_dict[status]


    # def check_finish_tag(self, job):
    #     job_tag_finished = job.job_hash + '_job_tag_finished'
    #     return self.context.check_file_exists(job_tag_finished)


class Lebesgue(DpCloudServer):
    pass



import os,sys,json,glob,shutil,uuid,time
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import dpdispatcher

class TestImportClasses(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_import_class_Machine(self):
        from dpdispatcher import Machine
        self.assertEqual(dpdispatcher.machine.Machine, Machine)
    
    def test_import_class_Resources(self):
        from dpdispatcher import Resources
        self.assertEqual(dpdispatcher.submission.Resources, Resources)
    
    def test_import_class_Submission(self):
        from dpdispatcher import Submission
        self.assertEqual(dpdispatcher.submission.Submission, Submission)

    def test_import_class_Task(self):
        from dpdispatcher import Task
        self.assertEqual(dpdispatcher.submission.Task, Task)



import os,sys,json,glob,shutil,uuid,time
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'shell'
from .context import LocalSession
from .context import LocalContext
from .context import Shell
from .context import JobStatus
from .context import my_file_cmp
from .context import setUpModule

class TestShell(unittest.TestCase) :
    def setUp(self) :
        os.makedirs('loc', exist_ok = True)
        os.makedirs('rmt', exist_ok = True)
        os.makedirs('loc/task0', exist_ok = True)
        os.makedirs('loc/task1', exist_ok = True)
        for ii in ['loc/task0', 'loc/task1']:
            with open(os.path.join(ii, 'test0'),'w') as fp:
                fp.write(str(uuid.uuid4()))
        work_profile = LocalSession({'work_path':'rmt'})
        self.ctx = LocalContext('loc', work_profile)
        self.shell = Shell(self.ctx)

    def tearDown(self):
        shutil.rmtree('loc')
        shutil.rmtree('rmt')
        if os.path.exists('dpgen.log'):
            os.remove('dpgen.log')
        if os.path.exists('run.sub'):
            os.remove('run.sub')
        if os.path.exists('run.sub.1'):
            os.remove('run.sub.1')

    def test_manual_cuda_devices(self):
        job_dirs = ['task0', 'task1']
        res = {'manual_cuda_devices': 3}
        ret = self.shell.sub_script(job_dirs, ['touch test1', 'touch test2'], res = res)
        with open('run.sub.gpu', 'w') as fp:
            fp.write(ret)        
            
    def test_manual_cuda_multiplicity(self):
        job_dirs = ['task0', 'task1', 'task2', 'task3']
        res = {'manual_cuda_devices': 2, 'manual_cuda_multiplicity': 2}
        ret = self.shell.sub_script(job_dirs, ['touch test1', 'touch test2'], res = res)
        with open('run.sub.gpu.multi', 'w') as fp:
            fp.write(ret)

    def test_gen_sub_script(self):
        job_dirs = ['task0', 'task1']
        self.shell.context.upload(job_dirs, ['test0'])
        ret = self.shell.sub_script(job_dirs, ['touch test1', 'touch test2'])
        with open('run.sub', 'w') as fp:
            fp.write(ret)
        ret1 = self.shell.sub_script(job_dirs, ['touch', 'touch'], args = [['test1 ', 'test1 '], ['test2 ', 'test2 ']])
        with open('run.sub.1', 'w') as fp:
            fp.write(ret1)
        time.sleep(1)
        my_file_cmp(self, 'run.sub.1', 'run.sub')
        # with open('run.sub', 'w') as fp:
        #     fp.write(ret)

    def test_sub_success(self) :
        job_dirs = ['task0', 'task1']
        self.shell.context.upload(job_dirs, ['test0'])
        self.shell.submit(job_dirs, ['touch test1', 'touch test2'])
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished  :
                break
            time.sleep(1)        
        self.assertTrue(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_0_finished')))
        self.assertTrue(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_1_finished')))
        self.assertTrue(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_0_finished')))
        self.assertTrue(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_1_finished')))
        self.assertTrue(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, '%s_tag_finished' % self.shell.context.job_uuid)))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test1')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test1')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test2')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test2')))


    def test_sub_scancel(self) :
        job_dirs = ['task0', 'task1']
        self.shell.context.upload(job_dirs, ['test0'])
        # sub        
        self.shell.submit(job_dirs, ['touch test1', 'sleep 10'])
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished  :
                raise RuntimeError('should not finished')
            if ret == JobStatus.running :
                # wait for file writing
                time.sleep(2)
                # kill job
                self.shell.context.kill(self.shell.proc)
                break
            time.sleep(1)
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_0_finished')))
        self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_1_finished')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_0_finished')))
        self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_1_finished')))
        self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, '%s_tag_finished' % self.shell.context.job_uuid)))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test1')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test1')))
        self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test2')))
        self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test2')))
        # sub restart
        self.shell.submit(job_dirs, ['rm test1', 'touch test2'], restart = True)
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished  :
                break
            time.sleep(1)
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_0_finished')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_1_finished')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_0_finished')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_1_finished')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, '%s_tag_finished' % self.shell.context.job_uuid)))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test1')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test1')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test2')))
        self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test2')))
        



import os
HTTP_TIME_OUT = 30

API_HOST = os.environ.get('DPDISPATCHER_LEBESGUE_API_HOST', "https://lebesgue.dp.tech")

ALI_STS_ENDPOINT = os.environ.get('DPDISPATCHER_LEBESGUE_ALI_STS_ENDPOINT', 'http://oss-cn-shenzhen.aliyuncs.com')
ALI_STS_BUCKET_NAME = os.environ.get('DPDISPATCHER_LEBESGUE_ALI_STS_BUCKET_NAME', "dpcloudserver")
ALI_OSS_BUCKET_URL = os.environ.get('DPDISPATCHER_LEBESGUE_ALI_OSS_BUCKET_URL', "https://dpcloudserver.oss-cn-shenzhen.aliyuncs.com/")
# ALI_OSS_BUCKET_URL = 'https://dpcloudserver.oss-cn-shenzhen.aliyuncs.com/


from dpdispatcher.machine import Machine
import time
import pathlib
from typing import List
from dargs import Argument

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.machine import script_command_template
# from dpdispatcher.submission import Resources

slurm_script_header_template="""\
#!/bin/bash -l
#SBATCH --parsable
{slurm_nodes_line}
{slurm_ntasks_per_node_line}
{slurm_number_gpu_line}
{slurm_partition_line}"""

class Slurm(Machine):
    def gen_script(self, job):
        slurm_script = super(Slurm, self).gen_script(job)
        return slurm_script

    def gen_script_header(self, job):
        resources = job.resources
        script_header_dict = {}
        script_header_dict['slurm_nodes_line']="#SBATCH --nodes {number_node}".format(number_node=resources.number_node)
        script_header_dict['slurm_ntasks_per_node_line']="#SBATCH --ntasks-per-node {cpu_per_node}".format(cpu_per_node=resources.cpu_per_node)
        custom_gpu_line = resources.kwargs.get("custom_gpu_line", None)
        if not custom_gpu_line:
            script_header_dict['slurm_number_gpu_line'] = "#SBATCH --gres=gpu:{gpu_per_node}".format(gpu_per_node=resources.gpu_per_node)
        else:
            script_header_dict['slurm_number_gpu_line'] = custom_gpu_line
        script_header_dict['slurm_partition_line']="#SBATCH --partition {queue_name}".format(queue_name=resources.queue_name)
        slurm_script_header = slurm_script_header_template.format(**script_header_dict)
        return slurm_script_header

    def do_submit(self, job, retry=0, max_retry=3):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + '_job_id'
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        ret, stdin, stdout, stderr = self.context.block_call('cd %s && %s %s' % (self.context.remote_root, 'sbatch', script_file_name))
        if ret != 0:
            err_str = stderr.read().decode('utf-8')
            if "Socket timed out on send/recv operation" in err_str or "Unable to contact slurm controller" in err_str:
                # server network error, retry 3 times
                if retry < max_retry:
                    dlog.warning("Get error code %d in submitting through ssh with job: %s . message: %s" %
                        (ret, job.job_hash, err_str))
                    dlog.warning("Sleep 60 s and retry submitting...")
                    # rest 60s
                    time.sleep(60)
                    return self.do_submit(job, retry=retry+1, max_retry=max_retry)
            elif "Job violates accounting/QOS policy" in err_str:
                # job number exceeds, skip the submitting
                return ''
            raise RuntimeError\
                ("status command squeue fails to execute\nerror message:%s\nreturn code %d\n" % (err_str, ret))
        subret = (stdout.readlines())
        # --parsable
        # Outputs only the job id number and the cluster name if present.
        # The values are separated by a semicolon. Errors will still be displayed.
        job_id = subret[0].split(";")[0].strip()
        self.context.write_file(job_id_name, job_id)        
        return job_id

    def default_resources(self, resources) :
        pass
    
    def check_status(self, job, retry=0, max_retry=3):
        job_id = job.job_id
        if job_id == '' :
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr \
            = self.context.block_call ('squeue -o "%.18i %.2t" -j ' + job_id)
        if (ret != 0) :
            err_str = stderr.read().decode('utf-8')
            if str("Invalid job id specified") in err_str :
                if self.check_finish_tag(job) :
                    dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            elif "Socket timed out on send/recv operation" in err_str or "Unable to contact slurm controller" in err_str:
                # retry 3 times
                if retry < max_retry:
                    dlog.warning("Get error code %d in checking status through ssh with job: %s . message: %s" %
                        (ret, job.job_hash, err_str))
                    dlog.warning("Sleep 60 s and retry checking...")
                    # rest 60s
                    time.sleep(60)
                    return self.check_status(job, retry=retry+1, max_retry=max_retry)
            raise RuntimeError("status command squeue fails to execute."
                "job_id:%s \n error message:%s\n return code %d\n" % (job_id, err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-1]
        if not (len(status_line.split()) == 2 and status_word.isupper()): 
            raise RuntimeError("Error in getting job status, " +
                              f"status_line = {status_line}, " + 
                              f"parsed status_word = {status_word}")
        if status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif status_word in ["R"] :
            return JobStatus.running
        elif status_word in ["CG"] :
            return JobStatus.completing
        elif status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown                    
        
    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        return self.context.check_file_exists(job_tag_finished)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.
        
        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_custom_gpu_line = "Custom GPU configuration, starting with #SBATCH"
        return [Argument("kwargs", dict, [
            Argument("custom_gpu_line", str, optional=True, default=None, doc=doc_custom_gpu_line)
        ], optional=True, doc="Extra arguments.")]


class SlurmJobArray(Slurm):
    """Slurm with job array enabled for multiple tasks in a job"""
    def gen_script_header(self, job):
        return super().gen_script_header(job) + "\n#SBATCH --array=0-%d" % (len(job.job_task_list)-1)

    def gen_script_command(self, job):
        resources = job.resources
        # SLURM_ARRAY_TASK_ID: 0 ~ n_jobs-1
        script_command = "case $SLURM_ARRAY_TASK_ID in\n"
        for ii, task in enumerate(job.job_task_list):
            command_env = ""
            command_env += self.gen_command_env_cuda_devices(resources=resources)

            task_tag_finished = task.task_hash + '_task_tag_finished'

            log_err_part = ""
            if task.outlog is not None:
                log_err_part += f"1>>{task.outlog} "
            if task.errlog is not None:
                log_err_part += f"2>>{task.errlog} "

            flag_if_job_task_fail = job.job_hash + '_flag_if_job_task_fail'
            single_script_command = script_command_template.format(
                flag_if_job_task_fail=flag_if_job_task_fail,
                command_env=command_env,
                task_work_path=pathlib.PurePath(task.task_work_path).as_posix(),
                command=task.command,
                task_tag_finished=task_tag_finished,
                log_err_part=log_err_part)
            script_command += f"{ii})\n"
            script_command += single_script_command
            script_command += self.gen_script_wait(resources=resources)
            script_command += "\n;;\n"
        script_command += "*)\nexit 1\n;;\nesac\n"
        return script_command

    def gen_script_end(self, job):
        # We cannot have a end script for job array
        # we may check task tag instead
        return ""

    def check_status(self, job, retry=0, max_retry=3):
        job_id = job.job_id
        if job_id == '' :
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr \
            = self.context.block_call ('squeue -h -o "%.18i %.2t" -j ' + job_id)
        if (ret != 0) :
            err_str = stderr.read().decode('utf-8')
            if str("Invalid job id specified") in err_str :
                if self.check_finish_tag(job) :
                    dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            elif "Socket timed out on send/recv operation" in err_str or "Unable to contact slurm controller" in err_str:
                # retry 3 times
                if retry < max_retry:
                    dlog.warning("Get error code %d in checking status through ssh with job: %s . message: %s" %
                        (ret, job.job_hash, err_str))
                    dlog.warning("Sleep 60 s and retry checking...")
                    # rest 60s
                    time.sleep(60)
                    return self.check_status(job, retry=retry+1, max_retry=max_retry)
            raise RuntimeError("status command squeue fails to execute."
                "job_id:%s \n error message:%s\n return code %d\n" % (job_id, err_str, ret))
        status_lines = stdout.read().decode('utf-8').split ('\n')[:-1]
        status = []
        for status_line in status_lines:
            status_word = status_line.split ()[-1]
            if not (len(status_line.split()) == 2 and status_word.isupper()): 
                raise RuntimeError("Error in getting job status, " +
                                f"status_line = {status_line}, " + 
                                f"parsed status_word = {status_word}")
            if status_word in ["PD","CF","S"] :
                status.append(JobStatus.waiting)
            elif status_word in ["R"] :
                status.append(JobStatus.running)
            elif status_word in ["CG"] :
                status.append(JobStatus.completing)
            elif status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
                status.append(JobStatus.finished)
            else :
                status.append(JobStatus.unknown)
        # running if any job is running
        if JobStatus.running in status:
            return JobStatus.running
        elif JobStatus.waiting in status:
            return JobStatus.waiting
        elif JobStatus.completing in status:
            return JobStatus.completing
        elif JobStatus.unknown in status:
            return JobStatus.unknown
        else:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else :
                return JobStatus.terminated

    def check_finish_tag(self, job):
        results = []
        for task in job.job_task_list:
            task_tag_finished = (pathlib.PurePath(task.task_work_path)/(task.task_hash + '_task_tag_finished')).as_posix()
            results.append(self.context.check_file_exists(task_tag_finished))
        return all(results)



import os,sys,json,glob,shutil,uuid,time
from socket import gaierror
import unittest
from unittest.mock import MagicMock, patch, PropertyMock


sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import LocalContext
from .context import BaseContext
from .context import PBS
from .context import JobStatus
from .context import LazyLocalContext, LocalContext, SSHContext
from .context import LSF, Slurm, PBS, Shell
from .context import Machine
from .context import dargs
from .context import DistributedShell, HDFSContext
from .sample_class import SampleClass
from dargs.dargs import ArgumentKeyError

class TestMachineInit(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
    
    def test_machine_serialize_deserialize(self):
        pbs = SampleClass.get_sample_pbs_local_context()
        self.assertEqual(pbs, Machine.deserialize(pbs.serialize()))

    def test_machine_load_from_dict(self):
        pbs = SampleClass.get_sample_pbs_local_context()
        self.assertEqual(pbs, PBS.load_from_dict(pbs.serialize()))



#%%
import os,sys,json,glob,shutil,uuid,time
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch, PropertyMock
#%%
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import LocalContext
# from .context import LocalSession
from .context import setUpModule
from .context import _identical_files
from .context import Task
from .context import get_file_md5
from .sample_class import SampleClass
# from .context import dpd

class TestIdFile(unittest.TestCase):
    def test_id(self) :
        with open('f0', 'w') as fp:
            fp.write('foo')
        with open('f1', 'w') as fp:
            fp.write('foo')
        self.assertTrue(_identical_files('f0', 'f1'))
        os.remove('f0')
        os.remove('f1')

    def test_diff(self) :
        with open('f0', 'w') as fp:
            fp.write('foo')
        with open('f1', 'w') as fp:
            fp.write('bar')
        self.assertFalse(_identical_files('f0', 'f1'))
        os.remove('f0')
        os.remove('f1')


class TestLocalContext(unittest.TestCase):
    def setUp(self):
        self.tmp_local_root = 'test_context_dir/'
        self.tmp_remote_root  = 'tmp_local_context_remote_root/'
        self.local_context = LocalContext(
            local_root=self.tmp_local_root,
            remote_root=self.tmp_remote_root
        )
    
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('tmp_local_context_remote_root/')

    def test_upload_non_exist(self):
        submission_hash = 'mock_hash_1'
        task1 = MagicMock(
            task_work_path='bct-1/',
            forward_files=['foo.py']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1],
            submission_hash=submission_hash)

        self.local_context.bind_submission(submission)

        with self.assertRaises(RuntimeError):
            self.local_context.upload(submission)

    def test_upload(self):
        submission_hash = 'mock_hash_2'
        task1 = MagicMock(
            task_work_path='bct-1/',
            forward_files=['input.lammps', 'conf.lmp']
        )
        task2 = MagicMock(
            task_work_path='bct-2/',
            forward_files=['input.lammps', 'conf.lmp']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1, task2],
            forward_common_files=['graph.pb'],
            submission_hash=submission_hash)

        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        check_file_list = [
            'bct-1/input.lammps',
            'bct-1/conf.lmp',
            'bct-2/input.lammps',
            'bct-2/conf.lmp',
            'graph.pb',
        ]
        for file in check_file_list:
            f1 = os.path.join(self.tmp_local_root, '0_md/', file)
            f2 = os.path.join(self.tmp_remote_root, submission_hash, file)
            self.assertEqual(get_file_md5(f1), get_file_md5(f2), msg=(f1,f2))

    def test_block_call(self):
        submission_hash = 'mock_hash_3'
        task1 = MagicMock(
            task_work_path='bct-1/',
            forward_files=['input.lammps', 'conf.lmp']
        )
        task2 = MagicMock(
            task_work_path='bct-2/',
            forward_files=['input.lammps', 'conf.lmp']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1, task2],
            submission_hash=submission_hash)
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)
        # time.sleep(0.01)
        code, stdin, stdout, stderr = self.local_context.block_call('ls')
        self.assertEqual(code, 0)
        self.assertEqual(stdout.readlines(), ['bct-1\n', 'bct-2\n'])

        code, stdin, stdout, stderr = self.local_context.block_call('ls a')
        self.assertEqual(code, 2)
        err_msg = stderr.read().decode('utf-8')
        self.assertTrue('ls: cannot access' in err_msg)
        self.assertTrue('No such file or directory\n' in err_msg)

    def test_call(self) :
        submission_hash = 'mock_hash_4'
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[],
            submission_hash=submission_hash)
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        proc = self.local_context.call('sleep 0.12')
        self.assertFalse(self.local_context.check_finish(proc))
        time.sleep(0.06)
        self.assertFalse(self.local_context.check_finish(proc))
        time.sleep(0.10)
        self.assertTrue(self.local_context.check_finish(proc))
        r,o,e=self.local_context.get_return(proc)
        self.assertEqual(r, 0)
        self.assertEqual(o.read(), b'')
        self.assertEqual(e.read(), b'')
        # not correct on centos7 aliyun ehpc
        # r,o,e=self.local_context.get_return(proc)
        # self.assertEqual(r, 0)
        # self.assertEqual(o, None)
        # self.assertEqual(e, None)

    def test_file(self):
        submission_hash = 'mock_hash_5'
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[],
            submission_hash=submission_hash)
        self.local_context.bind_submission(submission)
        self.local_context.upload(submission)

        self.assertFalse(self.local_context.check_file_exists('aaa'))
        tmp = str(uuid.uuid4())
        self.local_context.write_file('aaa', tmp)
        self.assertTrue(self.local_context.check_file_exists('aaa'))
        tmp1 = self.local_context.read_file('aaa')
        self.assertEqual(tmp, tmp1)

class TestLocalContextDownload(unittest.TestCase):
    # @classmethod
    # def setUpClass(cls):
    def setUp(self):
        shutil.copytree(src='test_context_dir/', 
            dst='tmp_local_context_download_dir/')
        os.makedirs('tmp_local_context_backfill_dir/0_md/bct-1/')
        os.makedirs('tmp_local_context_backfill_dir/0_md/bct-2/')

        self.tmp_local_root = 'tmp_local_context_backfill_dir'
        self.tmp_remote_root = 'tmp_local_context_download_dir'
        self.local_context = LocalContext(
            local_root=self.tmp_local_root,
            remote_root=self.tmp_remote_root
        )

    def tearDown(self):
        shutil.rmtree('tmp_local_context_download_dir/')
        shutil.rmtree('tmp_local_context_backfill_dir/')

    def test_download_trival(self):
        # submission_hash = 'mock_hash_2'
        task1 = MagicMock(
            task_work_path='bct-1/',
            backward_files=['input.lammps', 'conf.lmp']
        )
        task2 = MagicMock(
            task_work_path='bct-2/',
            backward_files=['input.lammps', 'conf.lmp']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1, task2],
            backward_common_files=['graph.pb'],
            submission_hash='0_md/')

        self.local_context.bind_submission(submission)

        check_file_list = [
            'bct-1/input.lammps',
            'bct-1/conf.lmp',
            'bct-2/input.lammps',
            'bct-2/conf.lmp',
            'graph.pb',
        ]

        self.local_context.download(submission)

        for file in check_file_list:
            f1 = os.path.join(self.tmp_local_root, '0_md/', file)
            self.assertTrue(os.path.isfile(f1))
            self.assertFalse(os.path.islink(f1))

    def test_download_check_exists(self):
        task1 = MagicMock(
            task_work_path='bct-1/',
            backward_files=['foo.py']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1],
            backward_common_files=['graph.pb'],
            submission_hash='0_md/')
        self.local_context.bind_submission(submission)
        with self.assertRaises(RuntimeError):
            self.local_context.download(
                submission, 
                check_exists=False)

    def test_download_mark_failure_tag(self):
        task1 = MagicMock(
            task_work_path='bct-1/',
            backward_files=['foo.py']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1],
            backward_common_files=['graph.pb'],
            submission_hash='0_md/')
        self.local_context.bind_submission(submission)
        # with self.assertRaises(RuntimeError):
        self.local_context.download(
            submission, 
            check_exists=True,
            mark_failure=True)
        
        tag_file = os.path.join(
            self.tmp_local_root,
            '0_md/',
            'bct-1/',
            'tag_failure_download_foo.py'
        )
        self.assertTrue(os.path.isfile(tag_file))
    
    def test_download_replace_old_files(self):
        task1 = MagicMock(
            task_work_path='bct-1/',
            backward_files=['input.lammps']
        )
        submission = MagicMock(work_base='0_md/',
            belonging_tasks=[task1],
            backward_common_files=['graph.pb'],
            submission_hash='0_md/')
        # with self.assertRaises(RuntimeError):
        target_file = os.path.join(
            self.tmp_local_root,
            '0_md/',
            'bct-1/',
            'input.lammps'
        )
        with open(target_file, 'w') as f:
            f.write("\n")
        md5_old = get_file_md5(target_file)
        self.local_context.bind_submission(submission)
        self.local_context.download(submission)
        md5_new = get_file_md5(target_file)
        self.assertNotEqual(md5_old, md5_new)




# 2开头的错误代码第二位代表错误等级
# 0. 严重错误; 1. 普通错误; 2. 规则错误; 3. 一般信息; 4. 未知错误
class RETCODE:
    OK                  = "0000" # 正常
    DBERR               = "2000" # 数据库异常
    THIRDERR            = "2001" # 第三方异常
    DATAERR             = "2002" # 数据异常
    IOERR               = "2003" # IO异常

    TOKENINVALID        = "2100" # 登陆错误
    PARAMERR            = "2101" # 参数错误
    USERERR             = "2102" # 用户异常
    ROLEERR             = "2103" # 权限错误
    PWDERR              = "2104" # 密码错误
    VERIFYERR           = "2105" # 验证错误

    REQERR              = "2200" # 请求错误

    NODATA              = "2300" # 无数据
    UNDERDEBUG          = "2301" # debug模式下无法使用

    UNKOWNERR           = "2400" # 未知错误


from dpdispatcher.base_context import BaseContext
import os,shutil,hashlib,tarfile
from glob import glob
from dpdispatcher import dlog
from dpdispatcher.hdfs_cli import HDFS

class HDFSContext(BaseContext) :
    def __init__(self,
                local_root,
                remote_root,
                remote_profile={},
                *args,
                **kwargs,
                ):

        assert(type(local_root) == str)
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.temp_remote_root = remote_root
        self.remote_profile = remote_profile

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict['local_root']
        remote_root = context_dict['remote_root']
        remote_profile = context_dict.get('remote_profile', {})
        instance = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile
        )
        return instance

    def get_job_root(self):
        return self.remote_root
    
    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = os.path.join(self.temp_remote_root, submission.submission_hash)

        HDFS.mkdir(self.remote_root)

    def _put_files(self,
                   files,
                   dereference=True):
        of = self.submission.submission_hash + '_upload.tgz'
        # local tar
        cwd = os.getcwd()
        os.chdir(self.local_root)
        if os.path.isfile(of) :
            os.remove(of)
        with tarfile.open(of, "w:gz", dereference=dereference) as tar:
            for ii in files :
                tar.add(ii)
        os.chdir(cwd)

        # trans
        from_f = os.path.join(self.local_root, of)
        HDFS.copy_from_local(os.path.join(self.local_root, of), self.remote_root)

        # clean up
        os.remove(from_f)

    def upload(self,
               submission,
               dereference=True):
        """ upload forward files and forward command files to HDFS root dir

        Parameters
        ----------
        submission : Submission class instance
            represents a collection of tasks, such as forward file names

        Returns
        -------
        none
        """

        cwd = os.getcwd()
        os.chdir(self.local_root)
        file_list = []

        for task in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, task.task_work_path)
            for ff in task.forward_files:
                abs_file_list = glob(os.path.join(local_job, ff))
                if not abs_file_list:
                    os.chdir(cwd)
                    raise RuntimeError('cannot find upload file ' + os.path.join(local_job, ff))
                rel_file_list = [os.path.relpath(ii, self.local_root) for ii in abs_file_list]
                file_list.extend(rel_file_list)

        local_job = self.local_root
        for fc in submission.forward_common_files:
            abs_file_list = glob(os.path.join(local_job, fc))
            if not abs_file_list:
                os.chdir(cwd)
                raise RuntimeError('cannot find upload file ' + os.path.join(local_job, fc))
            rel_file_list = [os.path.relpath(ii, self.local_root) for ii in abs_file_list]
            file_list.extend(rel_file_list)

        self._put_files(file_list, dereference=dereference)
        os.chdir(cwd)

    def download(self,
                 submission,
                 check_exists= False,
                 mark_failure = True,
                 back_error = False):
        """ download backward files from HDFS root dir

        Parameters
        ----------
        submission : Submission class instance
            represents a collection of tasks, such as backward file names

        Returns
        -------
        none
        """

        cwd = os.getcwd()

        # download all hdfs files to tmp dir
        os.chdir(self.local_root)
        gz_dir = os.path.join(self.local_root, 'tmp')
        if os.path.exists(gz_dir):
            shutil.rmtree(gz_dir, ignore_errors=True)
        os.mkdir('tmp')
        os.chdir(gz_dir)
        rfile_tgz = "%s/%s_*_download.tar.gz" % (self.remote_root, submission.submission_hash)
        lfile_tgz = "%s/tmp/" % (self.local_root)
        HDFS.copy_to_local(rfile_tgz, lfile_tgz)

        tgz_file_list = glob(os.path.join(self.local_root, "tmp/*_download.tar.gz"))
        for tgz in tgz_file_list:
            with tarfile.open(tgz, "r:gz") as tar:
                tar.extractall()
        os.chdir(self.local_root)

        for task in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, task.task_work_path)
            remote_job = os.path.join(gz_dir, task.task_work_path)
            flist = task.backward_files
            if back_error:
                errors = glob(os.path.join(remote_job, 'error*'))
                flist.extend(errors)
            for jj in flist:
                rfile = os.path.join(remote_job, jj)
                lfile = os.path.join(local_job, jj)

                if not os.path.exists(rfile):
                    if check_exists:
                        if mark_failure:
                            with open(os.path.join(self.local_root, ii, 'tag_failure_download_%s' % jj), 'w') as fp:
                                pass
                        else:
                            raise RuntimeError('do not find download file ' + rfile)
                    else:
                        raise RuntimeError('do not find download file ' + rfile)
                else:
                    if os.path.exists(lfile):
                        dlog.info('find existing %s, replacing by %s' % (lfile, rfile))
                        if os.path.isdir(lfile):
                            shutil.rmtree(lfile, ignore_errors=True)
                        elif os.path.isfile(lfile):
                            os.remove(lfile)
                    shutil.move(rfile, lfile)

        os.chdir(cwd)
        local_job = self.local_root
        remote_job = gz_dir
        flist = submission.backward_common_files
        if back_error:
            errors = glob(os.path.join(remote_job, 'error*'))
            flist.extend(errors)
        for jj in flist:
            rfile = os.path.join(remote_job, jj)
            lfile = os.path.join(local_job, jj)

            if not os.path.exists(rfile):
                if check_exists:
                    if mark_failure:
                        with open(os.path.join(self.local_root, ii, 'tag_failure_download_%s' % jj), 'w') as fp:
                            pass
                    else:
                        raise RuntimeError('do not find download file ' + rfile)
                else:
                    raise RuntimeError('do not find download file ' + rfile)
            else:
                if os.path.exists(lfile):
                    dlog.info('find existing %s, replacing by %s' % (lfile, rfile))
                    if os.path.isdir(lfile):
                        shutil.rmtree(lfile, ignore_errors=True)
                    elif os.path.isfile(lfile):
                        os.remove(lfile)
                shutil.move(rfile, lfile)

        # remove tmp dir
        shutil.rmtree(gz_dir, ignore_errors=True)

    def check_file_exists(self, fname):
        """ check whether the given file exists, often used in checking whether the belonging job has finished

         Parameters
         ----------
         fname : string
             file name to be checked

         Returns
         -------
         status: boolean
         """
        return HDFS.exists(os.path.join(self.remote_root, fname))

    def clean(self):
        HDFS.remove(self.remote_root)

    def write_file(self, fname, write_str):
        local_file = os.path.join("/tmp/", fname)
        with open(local_file, 'w') as fp:
            fp.write(write_str)
        HDFS.copy_from_local(local_file, os.path.join(self.remote_root, fname))
        return local_file

    def read_file(self, fname):
        return HDFS.read_hdfs_file(os.path.join(self.remote_root, fname))

    def check_file_exists(self, fname):
        return HDFS.exists(os.path.join(self.remote_root, fname))

    def kill(self, job_id):
        pass



from dpdispatcher import Resources, Task, Machine
from dargs import ArgumentEncoder
import json

resources_dargs = Resources.arginfo()
with open('dpdispatcher-resources.json', 'w') as f:
    json.dump(resources_dargs, f, cls=ArgumentEncoder)

machine_dargs = Machine.arginfo()
with open('dpdispatcher-machine.json', 'w') as f:
    json.dump(machine_dargs, f, cls=ArgumentEncoder)

task_dargs = Task.arginfo()
with open('dpdispatcher-task.json', 'w') as f:
    json.dump(task_dargs, f, cls=ArgumentEncoder)




import os,sys,json,glob,shutil,uuid,getpass
import unittest
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import SSHContext, SSHSession
from .context import setUpModule

class TestSSHContext(unittest.TestCase):
    def setUp(self) :
        os.makedirs('loc', exist_ok = True)
        os.makedirs('loc/task0', exist_ok = True)
        os.makedirs('loc/task1', exist_ok = True)
        for ii in ['loc/task0', 'loc/task1']:
            with open(os.path.join(ii, 'test0'),'w') as fp:
                fp.write(str(uuid.uuid4()))
            with open(os.path.join(ii, 'test1'),'w') as fp:
                fp.write(str(uuid.uuid4()))
            os.makedirs(os.path.join(ii, 'dir0'), exist_ok = True)
            with open(os.path.join(ii, 'dir0', 'test2'),'w') as fp:
                fp.write(str(uuid.uuid4()))
        os.makedirs('rmt', exist_ok = True)
        try :
            self.ssh_session = SSHSession({'hostname' : 'localhost',
                                           'port': 22,
                                           'username' : getpass.getuser(),
                                           'work_path' : os.path.join(os.getcwd(), 'rmt')})
        except:
            # for tianhe-2
            try:
                self.ssh_session = SSHSession({'hostname' : 'localhost',
                                            'port': 5566,
                                            'username' : getpass.getuser(),
                                            'work_path' : os.path.join(os.getcwd(), 'rmt')})
            except:
                self.skipTest("Network error")
        self.job = SSHContext('loc', self.ssh_session)
        self.job1 = SSHContext('loc', self.ssh_session, job_uuid = self.job.job_uuid)
    
    def tearDown(self):
        self.job.close()
        self.job1.close()
        shutil.rmtree('loc')
        shutil.rmtree('rmt')

    def test_upload(self) :
        tasks = ['task0', 'task1']
        files = ['test0', 'test1']
        self.job.upload(tasks, files)
        for ii in tasks :
            for jj in files :
                locf = os.path.join('loc', ii, jj)
                rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj)
                with open(locf) as fp:
                    locs = fp.read()
                with open(rmtf) as fp:
                    rmts = fp.read()
                self.assertEqual(locs, rmts)
        self.job.upload(tasks, ['dir0'])
        for ii in tasks :
            for jj in ['dir0'] :
                for kk in ['test2'] :
                    locf = os.path.join('loc', ii, jj, kk)
                    rmtf = os.path.join('rmt', self.job.job_uuid, ii, jj, kk)
                    with open(locf) as fp:
                        locs = fp.read()
                    with open(rmtf) as fp:
                        rmts = fp.read()
                    self.assertEqual(locs, rmts)


    def test_donwload(self):
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test4', 'test5'] :
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # generate extra donwload dirs and files
        for ii in tasks :
            for jj in ['dir1'] :
                os.makedirs(os.path.join('rmt',self.job.job_uuid,ii,jj), exist_ok=False)
                for kk in ['test6']:
                    with open(os.path.join('rmt',self.job.job_uuid,ii,jj,kk), 'w') as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
                        record_uuid.append(tmp)                        
        # donwload
        files = ['test4', 'test5', 'dir1']
        self.job.download(tasks, files)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test4', 'test5'] :
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                for kk in ['test6']:
                    with open(os.path.join('loc',ii,jj,kk), 'r') as fp:
                        tmp = fp.read()
                        self.assertEqual(tmp, record_uuid[cc])
                        cc += 1


    def test_donwload_check_mark(self):
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test6', 'test7'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6'):
                    continue
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ['test6', 'test7', 'dir1']
        self.job.download(tasks, files, check_exists = True, mark_failure = True)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test6', 'test7'] :
                if (ii == 'task1' and jj == 'test7') or \
                   (ii == 'task0' and jj == 'test6') :
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, jj))
                    self.assertTrue(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)), 
                                    msg = 'failed to find ' + os.path.join('loc', ii, 'tag_failure_download_%s' % jj))
                    continue
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)))
                self.assertTrue(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)))


    def test_donwload_check_nomark(self):
        tasks = ['task0', 'task1']
        self.job.upload(tasks, ['test0', 'dir0'])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks :
            for jj in ['test6', 'test7'] :
                if ii == 'task1' and jj == 'test7' :
                    continue
                if ii == 'task0' and jj == 'test6' :
                    continue
                with open(os.path.join('rmt',self.job.job_uuid,ii,jj), 'w') as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ['test6', 'test7', 'dir1']
        self.job.download(tasks, files, check_exists = True, mark_failure = False)
        # check dlded
        cc = 0
        for ii in tasks :
            for jj in ['test6', 'test7'] :
                if ii == 'task1' and jj == 'test7' :
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, jj))
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, 'tag_failure_download_%s' % jj))
                    continue
                if ii == 'task0' and jj == 'test6' :
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, jj))
                    self.assertFalse(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)), 
                                     msg = 'found ' + os.path.join('loc', ii, 'tag_failure_download_%s' % jj))
                    continue
                with open(os.path.join('loc',ii,jj), 'r') as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks :
            for jj in ['dir1'] :
                self.assertFalse(os.path.exists(os.path.join('loc', ii, jj)))
                self.assertFalse(os.path.exists(os.path.join('loc', ii, 'tag_failure_download_%s' % jj)))

    def test_block_call(self) :
        tasks = ['task0', 'task1']
        files = ['test0', 'test1']
        self.job.upload(tasks, files)        
        # ls
        code, stdin, stdout, stderr = self.job.block_call('ls')
        self.assertEqual(stdout.read(), b'task0\ntask1\n')
        self.assertEqual(code, 0)
        code, stdin, stdout, stderr = self.job.block_call('ls')
        self.assertEqual(stdout.readlines(), ['task0\n','task1\n'])
        code, stdin, stdout, stderr = self.job.block_call('ls a')
        self.assertEqual(code, 2)
        # self.assertEqual(stderr.read().decode('utf-8'), 'ls: cannot access a: No such file or directory\n')
        err_msg = stderr.read().decode('utf-8')
        self.assertTrue('ls: cannot access' in err_msg)
        self.assertTrue('No such file or directory\n' in err_msg)
                        
    def test_block_checkcall(self) :
        tasks = ['task0', 'task1']
        files = ['test0', 'test1']
        self.job.upload(tasks, files)        
        # ls
        stdin, stdout, stderr = self.job.block_checkcall('ls')
        self.assertEqual(stdout.read(), b'task0\ntask1\n')
        stdin, stdout, stderr = self.job.block_checkcall('ls')
        self.assertEqual(stdout.readlines(), ['task0\n','task1\n'])
        with self.assertRaises(RuntimeError):
            stdin, stdout, stderr = self.job.block_checkcall('ls a')

    def test_file(self) :
        self.assertFalse(self.job.check_file_exists('aaa'))
        tmp = str(uuid.uuid4())
        self.job.write_file('aaa', tmp)
        self.assertTrue(self.job.check_file_exists('aaa'))
        tmp1 = self.job.read_file('aaa')
        self.assertEqual(tmp, tmp1)





#!/usr/bin/env python
# coding: utf-8
# %%
import uuid

from dargs.dargs import Argument
from dpdispatcher.base_context import BaseContext
from typing import List
import os
from dpdispatcher import dlog
# from dpdispatcher.submission import Machine
# from . import dlog
from .dpcloudserver.api import API
from .dpcloudserver import zip_file
import shutil
import tqdm
# from zip_file import zip_files
DP_CLOUD_SERVER_HOME_DIR = os.path.join(
    os.path.expanduser('~'),
    '.dpdispatcher/',
    'dp_cloud_server/'
)
ENDPOINT = 'http://oss-cn-shenzhen.aliyuncs.com'
BUCKET_NAME = 'dpcloudserver'

class DpCloudServerContext(BaseContext):
    def __init__ (self,
        local_root,
        remote_root=None,
        remote_profile={},
        *args,
        **kwargs,
    ):
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.remote_profile = remote_profile
        email = remote_profile.get("email", None)
        password = remote_profile.get('password')
        if email is None:
            raise ValueError("can not find email in remote_profile, please check your machine file.")
        if password is None:
            raise ValueError("can not find password in remote_profile, please check your machine file.")
        self.api = API(email, password)

        os.makedirs(DP_CLOUD_SERVER_HOME_DIR, exist_ok=True)

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict['local_root']
        remote_root = context_dict.get('remote_root', None)
        remote_profile = context_dict.get('remote_profile', {})

        dp_cloud_server_context = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile
        )
        return dp_cloud_server_context

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = '$(pwd)'

        self.submission_hash = submission.submission_hash

        self.machine = submission.machine

        # def zip_files(self, submission):
        #     file_uuid = uuid.uuid1().hex
        # oss_task_dir = os.path.join()

    def _gen_oss_path(self, job, zip_filename):
        if hasattr(job, 'upload_path') and job.upload_path:
            return job.upload_path
        else:
            program_id = self.remote_profile.get('program_id')
            if program_id is None:
                program_id = 0
            uid = uuid.uuid4()
            path = os.path.join("program", str(program_id), str(uid), zip_filename)
            setattr(job, 'upload_path', path)
            return path

    def upload(self, submission):
        # oss_task_dir = os.path.join('%s/%s/%s.zip' % ('indicate', file_uuid, file_uuid))
        # zip_filename = submission.submission_hash + '.zip'
        # oss_task_zip = 'indicate/' + submission.submission_hash + '/' + zip_filename

        # zip_path = "/home/felix/workplace/22_dpdispatcher/dpdispatcher-yfb/dpdispatcher/dpcloudserver/t.txt"
        # zip_path = self.local_root
        bar_format = "{l_bar}{bar}| {n:.02f}/{total:.02f} %  [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        job_to_be_uploaded = []
        result = None
        dlog.info("checking all job has been uploaded")
        for job in submission.belonging_jobs:
            if not self.api.check_job_has_uploaded(job.job_id):
                job_to_be_uploaded.append(job)
        if len(job_to_be_uploaded) == 0:
            dlog.info("all job has been uploaded, continue")
            return result
        for job in tqdm.tqdm(job_to_be_uploaded, desc="Uploading to Lebesgue", bar_format=bar_format):
            self.machine.gen_local_script(job)
            zip_filename = job.job_hash + '.zip'
            oss_task_zip = self._gen_oss_path(job, zip_filename)
            zip_task_file = os.path.join(self.local_root, zip_filename)

            upload_file_list = [job.script_file_name, ]
            upload_file_list.extend(submission.forward_common_files)

            for task in job.job_task_list:
                for file in task.forward_files:
                    upload_file_list.append(
                        os.path.join(
                            task.task_work_path, file
                        )
                    )

            upload_zip = zip_file.zip_file_list(
                self.local_root,
                zip_task_file,
                file_list=upload_file_list
            )
            result = self.api.upload(oss_task_zip, upload_zip, ENDPOINT, BUCKET_NAME)
            self._backup(self.local_root, upload_zip)
        return result
        # return oss_task_zip
        # api.upload(self.oss_task_dir, zip_task_file)

    def download(self, submission):
        jobs = submission.belonging_jobs
        job_hashs = {}
        group_id = None
        job_infos = {}
        for job in jobs:
            ids = job.job_id.split(":job_group_id:")
            jid, gid = int(ids[0]), int(ids[1])
            job_hashs[jid] = job.job_hash
            group_id = gid
        if group_id is not None:
            job_result = self.api.get_tasks_list(group_id)
            for each in job_result:
                if 'result_url' in each and each['result_url'] != '' and each['status'] == 2:
                    job_hash = ''
                    if each['task_id'] not in job_hashs:
                        dlog.info(f"find unexpect job_hash, but task {each['task_id']} still been download.")
                        dlog.debug(str(job_hashs))
                        job_hash = str(each['task_id'])
                    else:
                        job_hash = job_hashs[each['task_id']]
                    job_infos[job_hash] = each
        bar_format = "{l_bar}{bar}| {n:.02f}/{total:.02f} %  [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        for job_hash, info in tqdm.tqdm(job_infos.items(), desc="Validating download file from Lebesgue", bar_format=bar_format):
            result_filename = job_hash + '_back.zip'
            target_result_zip = os.path.join(self.local_root, result_filename)
            if self._check_if_job_has_already_downloaded(target_result_zip, self.local_root):
                continue
            self.api.download_from_url(info['result_url'], target_result_zip)
            zip_file.unzip_file(target_result_zip, out_dir=self.local_root)
            self._backup(self.local_root, target_result_zip)
        self._clean_backup(self.local_root, keep_backup=self.remote_profile.get('keep_backup', True))
        return True

    def _check_if_job_has_already_downloaded(self, target, local_root):
        backup_file_location = os.path.join(local_root, 'backup', os.path.split(target)[1])
        if os.path.exists(backup_file_location):
            return True
        else:
            return False

    def _backup(self, local_root, target):
        try:
            # move to backup directory
            os.makedirs(os.path.join(local_root, 'backup'), exist_ok=True)
            shutil.move(target,
                        os.path.join(local_root, 'backup', os.path.split(target)[1]))
        except (OSError, shutil.Error) as e:
            dlog.exception("unable to backup file, " + str(e))

    def _clean_backup(self, local_root, keep_backup=True):
        if not keep_backup:
            dir_to_be_removed = os.path.join(local_root, 'backup')
            if os.path.exists(dir_to_be_removed):
                shutil.rmtree(dir_to_be_removed)

    def write_file(self, fname, write_str):
        result = self.write_home_file(fname, write_str)
        return result

    def write_local_file(self, fname, write_str):
        local_filename = os.path.join(self.local_root, fname)
        with open(local_filename, 'w') as f:
            f.write(write_str)
        return local_filename

    def read_file(self, fname):
        result = self.read_home_file(fname)
        return result

    def write_home_file(self, fname, write_str):
        # os.makedirs(self.remote_root, exist_ok = True)
        with open(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname), 'w') as fp:
            fp.write(write_str)
        return True

    def read_home_file(self, fname):
        with open(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname), 'r') as fp:
            ret = fp.read()
        return ret

    def check_file_exists(self, fname):
        result = self.check_home_file_exits(fname)
        return result

    def check_home_file_exits(self, fname):
        return os.path.isfile(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname))

    def clean(self):
        submission_file_name = "{submission_hash}.json".format(
            submission_hash=self.submission.submission_hash
        )
        submission_json = os.path.join(
            DP_CLOUD_SERVER_HOME_DIR,
            submission_file_name
        )
        os.remove(submission_json)
        return True

    # def get_return(self, cmd_pipes):
    #     if not self.check_finish(cmd_pipes):
    #         return None, None, None
    #     else :
    #         retcode = cmd_pipes['stdout'].channel.recv_exit_status()
    #         return retcode, cmd_pipes['stdout'], cmd_pipes['stderr']

    def kill(self, cmd_pipes) :
        pass

    @classmethod
    def machine_subfields(cls) -> List[Argument]:
        """Generate the machine subfields.
        
        Returns
        -------
        list[Argument]
            machine subfields
        """
        doc_remote_profile = "The information used to maintain the connection with remote machine."
        return [Argument("remote_profile", dict, [
            Argument("email", str, optional=False, doc="Email"),
            Argument("password", str, optional=False, doc="Password"),
            Argument("program_id", int, optional=False, doc="Program ID"),
            Argument("input_data", dict, [
                Argument("job_name", str, optional=True, doc="Job name"),
                Argument("image_name", str, optional=True, doc="Name of the image which run the job, optional "
                                                               "when platform is not ali/oss."),
                Argument("disk_size", str, optional=True, doc="disk size (GB), optional "
                                                              "when platform is not ali/oss."),
                Argument("scass_type", str, optional=False, doc="machine configuration."),
                Argument("platform", str, optional=False, doc="Job run in which platform."),
                Argument("log_file", str, optional=True, doc="location of log file."),
                Argument('checkpoint_files', [str, list], optional=True, doc="location of checkpoint files when "
                                                                                  "it is list type. record file "
                                                                                  "changes when it is string value "
                                                                                  "'sync_files'"),
                Argument('checkpoint_time', int, optional=True, default=15, doc='interval of checkpoint data been '
                                                                                'stored minimum 15.'),
                Argument('backward_files', list, optional=True, doc='which files to be uploaded to remote '
                                                                         'resources. Upload all the files when it is '
                                                                         'None or empty.')
            ], optional=False, doc="Configuration of job"),
        ], doc=doc_remote_profile)]


class LebesgueContext(DpCloudServerContext):
    pass

#%%



import logging
import os, sys
import warnings

ROOT_PATH=__path__[0]
dlog = logging.getLogger(__name__)
dlog.propagate = False
dlog.setLevel(logging.INFO)
try:
    dlogf = logging.FileHandler(os.getcwd()+os.sep+'dpdispatcher'+'.log')
except PermissionError:
    warnings.warn(f"dpdispatcher.log meet permission error. redirect the log to ~/dpdispatcher.log")
    dlogf = logging.FileHandler(os.path.join(os.path.expanduser('~'),'dpdispatcher.log'))

# dlogf = logging.FileHandler('./'+os.sep+SHORT_CMD+'.log')
# dlogf = logging.FileHandler(os.path.join(os.environ['HOME'], SHORT_CMD+'.log'))
# dlogf = logging.FileHandler(os.path.join(os.path.expanduser('~'), SHORT_CMD+'.log'))
# dlogf = logging.FileHandler(os.path.join("/tmp/", SHORT_CMD+'.log'))
dlogf_formatter=logging.Formatter('%(asctime)s - %(levelname)s : %(message)s')
# dlogf_formatter=logging.Formatter('%(asctime)s - %(name)s - [%(filename)s:%(funcName)s - %(lineno)d ] - %(levelname)s \n %(message)s')
dlogf.setFormatter(dlogf_formatter)
dlog.addHandler(dlogf)

dlog_stdout=logging.StreamHandler(sys.stdout)
dlog_stdout.setFormatter(dlogf_formatter)
dlog.addHandler(dlog_stdout)

__author__    = "DeepModeling Team"
__copyright__ = "Copyright 2019"
__status__    = "Development"
try:
    from ._version import version as __version__
except ImportError:
    __version__ = 'unkown'
try:
    from ._date import date as __date__
except ImportError:
    __date__ = 'unkown'

from .submission import Submission
from .submission import Task
from .submission import Job
from .submission import Resources
from .slurm import Slurm
from .pbs import PBS
from .pbs import Torque
from .shell import Shell
from .lsf import LSF
from .dp_cloud_server import DpCloudServer, Lebesgue
from .distributed_shell import DistributedShell
from .machine import Machine

from .lazy_local_context import LazyLocalContext
from .local_context import LocalContext
from .ssh_context import SSHContext
from .dp_cloud_server_context import DpCloudServerContext, LebesgueContext
from .hdfs_context import HDFSContext

def info():
    """
        Show basic information about """+NAME+""", its location and version.
    """

    print('DeepModeling\n------------')
    print('Version: ' + __version__)
    print('Date:    ' + __date__)
    print('Path:    ' + ROOT_PATH)
    print('')
    print('Dependency')
    print('------------')
    for modui in ['psutil', 'paramiko', 'dargs', 'oss2']:
        try:
            mm = __import__(modui)
            print('%10s %10s   %s' % (modui, mm.__version__, mm.__path__[0]))
        except ImportError:
            print('%10s %10s Not Found' % (modui, ''))
    print()




import os
import json
import time

try:
    import oss2
    from oss2 import SizedFileAdapter, determine_part_size
    from oss2.models import PartInfo
except ImportError:
    pass
import requests
from urllib.parse import urljoin
from dpdispatcher import dlog

from .retcode import RETCODE
from .config import HTTP_TIME_OUT, API_HOST


class API:
    def __init__(self, email, password):
        self._login_data = {"password": password, 'email': email}
        self._token = None
        self.refresh_token()
        return

    def get(self, url, params, retry=0):
        headers = {'Authorization': "jwt " + self._token}
        ret = None
        for retry_count in range(3):
            try:
                ret = requests.get(
                    urljoin(API_HOST, url),
                    params=params,
                    timeout=HTTP_TIME_OUT,
                    headers=headers
                )
            except Exception as e:
                dlog.error(f"request error {e}")
                continue
            if ret.ok:
                break
            else:
                dlog.error(f"request error status_code:{ret.status_code} reason: {ret.reason} body: \n{ret.text}")
                time.sleep(retry_count * 10)
        if ret is None:
            raise ConnectionError("request fail")
        # print(url,'>>>', params, '<<<', ret.text)
        ret.raise_for_status()
        ret = json.loads(ret.text)
        if ret['code'] == RETCODE.TOKENINVALID and retry <= 3:
            dlog.info("debug: token expire, refresh token")
            if self._login_data is not None:
                self.refresh_token()
                ret = self.get(url, params, retry=retry + 1)
                return ret
        if ret['code'] != RETCODE.OK:
            raise ValueError(f"{url} Error: {ret['code']} {ret['message']}")
        return ret['data']

    def post(self, url, params, retry=0):
        headers = {'Authorization': "jwt " + self._token}
        ret = None
        for retry_count in range(3):
            try:
                ret = requests.post(
                    urljoin(API_HOST, url),
                    json=params,
                    timeout=HTTP_TIME_OUT,
                    headers=headers
                )
            except Exception as e:
                dlog.error(f"request error {e}")
                continue
            if ret.ok:
                break
            else:
                dlog.error(f"request error status_code:{ret.status_code} reason: {ret.reason} body: \n{ret.text}")
                time.sleep(retry_count)
        if ret is None:
            raise ConnectionError("request fail")
        ret.raise_for_status()
        ret = json.loads(ret.text)
        # print(url,'>>>', params, '<<<', ret.text)
        if ret['code'] == RETCODE.TOKENINVALID and retry <= 3:
            dlog.info("debug: token expire, refresh token")
            if self._login_data is not None:
                self.refresh_token()
                ret = self.post(url, params, retry=retry + 1)
                return ret
        if ret['code'] != RETCODE.OK:
            raise ValueError(f"{url} Error: {ret['code']} {ret['message']}")
        return ret['data']

    def refresh_token(self):
        url = '/account/login'
        ret = requests.post(
            urljoin(API_HOST, url),
            json=self._login_data,
            timeout=HTTP_TIME_OUT,
        )
        dlog.debug(f"debug: login ret:{ret}")
        ret = json.loads(ret.text)
        if ret['code'] != RETCODE.OK:
            raise ValueError(f"{url} Error: {ret['code']} {ret['message']}")
        self._token = ret['data']['token']

    def _get_oss_bucket(self, endpoint, bucket_name):
        #  res = get("/tools/sts_token", {})
        res = self.get("/data/get_sts_token", {})
        # print('debug>>>>>>>>>>>>>', res)
        dlog.debug(f"debug: _get_oss_bucket: res:{res}")
        auth = oss2.StsAuth(
            res['AccessKeyId'],
            res['AccessKeySecret'],
            res['SecurityToken']
        )
        return oss2.Bucket(auth, endpoint, bucket_name)

    def download(self, oss_file, save_file, endpoint, bucket_name):
        bucket = self._get_oss_bucket(endpoint, bucket_name)
        dlog.debug(f"debug: download: oss_file:{oss_file}; save_file:{save_file}")
        bucket.get_object_to_file(oss_file, save_file)
        return save_file

    def download_from_url(self, url, save_file):
        ret = None
        for retry_count in range(3):
            try:
                ret = requests.get(
                    url,
                    headers={'Authorization': "jwt " + self._token},
                    stream=True
                )
            except Exception as e:
                dlog.error(f"request error {e}")
                continue
            if ret.ok:
                break
            else:
                dlog.error(f"request error status_code:{ret.status_code} reason: {ret.reason} body: \n{ret.text}")
                time.sleep(retry_count)
                ret = None
        if ret is not None:
            ret.raise_for_status()
            with open(save_file, 'wb') as f:
                for chunk in ret.iter_content(chunk_size=8192):
                    f.write(chunk)
            ret.close()


    def upload(self, oss_task_zip, zip_task_file, endpoint, bucket_name):
        dlog.debug(f"debug: upload: oss_task_zip:{oss_task_zip}; zip_task_file:{zip_task_file}")
        bucket = self._get_oss_bucket(endpoint, bucket_name)
        total_size = os.path.getsize(zip_task_file)
        part_size = determine_part_size(total_size, preferred_size=1000 * 1024)
        upload_id = bucket.init_multipart_upload(oss_task_zip).upload_id
        parts = []
        with open(zip_task_file, 'rb') as fileobj:
            part_number = 1
            offset = 0
            while offset < total_size:
                num_to_upload = min(part_size, total_size - offset)
                result = bucket.upload_part(oss_task_zip, upload_id, part_number,
                                            SizedFileAdapter(fileobj, num_to_upload))
                parts.append(PartInfo(part_number, result.etag))
                offset += num_to_upload
                part_number += 1
        # result = bucket.complete_multipart_upload(oss_task_zip, upload_id, parts)
        result = bucket.complete_multipart_upload(oss_task_zip, upload_id, parts)
        # print('debug:upload_result:', result, dir())
        return result


    def job_create(self, job_type, oss_path, input_data, program_id=None, group_id=None):
        post_data = {
            'job_type': job_type,
            'oss_path': oss_path,
        }
        if program_id is not None:
            post_data["program_id"] = program_id
        if group_id is not None:
            post_data["job_group_id"] = group_id
        if input_data.get('command') is not None:
            post_data["cmd"] = input_data.get('command')
        if input_data.get('backward_files') is not None:
            post_data["out_files"] = input_data.get('backward_files')
        for k, v in input_data.items():
            post_data[k] = v
        ret = self.post('/data/v2/insert_job', post_data)
        group_id = ret.get('job_group_id')
        return ret['job_id'], group_id

    def get_jobs(self, page=1, per_page=10):
        ret = self.get(
            '/data/jobs',
            {
                'page': page,
                'per_page': per_page,
            }
        )
        return ret['items']

    def get_tasks(self, job_id, group_id, page=1, per_page=10):
        ret = self.get(
            f'data/job/{group_id}/tasks',
            {
                'page': page,
                'per_page': per_page,
            }
        )
        for each in ret['items']:
            if job_id == each["task_id"]:
                return each
        if len(ret['items']) != 0:
            return self.get_tasks(job_id, group_id, page=page + 1)
        return None

    def get_tasks_list(self, group_id, per_page=30):
        result = []
        page = 0
        while True:
            ret = self.get(
                f'data/job/{group_id}/tasks',
                {
                    'page': page,
                    'per_page': per_page,
                }
            )
            if len(ret['items']) == 0:
                break
            for each in ret['items']:
                result.append(each)
            page += 1
        return result

    def check_job_has_uploaded(self, job_id):
        try:
            if not job_id:
                return False
            if 'job_group_id' in job_id:
                ids = job_id.split(":job_group_id:")
                job_id, _ = int(ids[0]), int(ids[1])
            ret = self.get(f'data/job/{job_id}', {})
            if len(ret) == 0:
                return False
            if ret.get('input_data'):
                return True
            else:
                return False
        except ValueError as e:
            dlog.error(e)
            return False

    def get_job_result_url(self, job_id):
        try:
            if not job_id:
                return None
            if 'job_group_id' in job_id:
                ids = job_id.split(":job_group_id:")
                job_id, _ = int(ids[0]), int(ids[1])
            ret = self.get(f'data/job/{job_id}', {})
            if 'result_url' in ret and len(ret['result_url']) != 0:
                return ret.get('result_url')
            else:
                return None
        except ValueError as e:
            dlog.error(e)
            return None

# %%



import os
import sys
import json
from dpdispatcher.submission import Submission, Job, Task, Resources
from dpdispatcher.machine import Machine

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# task_need_resources has no effect
with open("jsons/machine_lsf.json", 'r') as f:
    mdata = json.load(f)

machine = Machine.load_from_dict(mdata['machine'])
resources = Resources.load_from_dict(mdata['resources'])

submission = Submission(
    work_base='0_md/',
    machine=machine,
    resources=resources,
    forward_common_files=['graph.pb'],
    backward_common_files=[]
)

task1 = Task(
    command='lmp -i input.lammps',
    task_work_path='bct-1/',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps']
)
task2 = Task(
    command='lmp -i input.lammps',
    task_work_path='bct-2/',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps']
)
task3 = Task(
    command='lmp -i input.lammps',
    task_work_path='bct-3/',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps']
)
task4 = Task(
    command='lmp -i input.lammps',
    task_work_path='bct-4/',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps']
)
submission.register_task_list([task1, task2, task3, task4, ])
submission.run_submission(clean=True)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import subprocess
from datetime import date
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'DPDispatcher'
copyright = '2020-%d, Deep Modeling' % date.today().year
author = 'Deep Modeling'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'recommonmark',
    "sphinx_rtd_theme",
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'numpydoc',
    'sphinx.ext.autosummary'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = []

autodoc_default_flags = ['members']
autosummary_generate = True
master_doc = 'index'

def run_apidoc(_):
    from sphinx.ext.apidoc import main
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    module = os.path.join(cur_dir, "..", "dpdispatcher")
    main(['-M', '--tocfile', 'api', '-H', 'DPDispatcher API', '-o', os.path.join(cur_dir, "api"), module, '--force'])

def generate_train_input(app):
    subprocess.check_output((sys.executable, os.path.join("..", "scripts", "script_gen_dargs_docs.py")), universal_newlines=True)

def setup(app):
    app.connect('builder-inited', run_apidoc)
    app.connect('builder-inited', generate_train_input)


intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
}



from dargs import Argument
from typing import List

from dpdispatcher import dlog

class BaseContext(object):
    subclasses_dict = {}
    options = set()
    def __new__(cls, *args, **kwargs):
        if cls is BaseContext:
            subcls = cls.subclasses_dict[kwargs['context_type']]
            instance = subcls.__new__(subcls, *args, **kwargs)
        else:
            instance = object.__new__(cls)
        return instance

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses_dict[cls.__name__]=cls
        cls.subclasses_dict[cls.__name__.lower()]=cls
        cls.subclasses_dict[cls.__name__.replace("Context", "")]=cls
        cls.subclasses_dict[cls.__name__.lower().replace("context", "")]=cls
        cls.options.add(cls.__name__)

    @classmethod
    def load_from_dict(cls, context_dict):
        context_type = context_dict['context_type']
        # print("debug778:context_type", cls.subclasses_dict, context_type)
        try:
            context_class = cls.subclasses_dict[context_type]
        except KeyError as e:
            dlog.error(f"KeyError:context_type; context_type:{context_type}; cls.subclasses_dict:{cls.subclasses_dict}")
            raise e
        context = context_class.load_from_dict(context_dict)
        return context

    def bind_submission(self, submission):
        self.submission = submission

    def upload(self, submission):
        raise NotImplementedError('abstract method')

    def download(self, 
                submission,
                check_exists = False,
                mark_failure = True,
                back_error=False):
        raise NotImplementedError('abstract method')

    def clean(self):
        raise NotImplementedError('abstract method')

    def write_file(self, fname, write_str):
        raise NotImplementedError('abstract method')

    def read_file(self, fname):
        raise NotImplementedError('abstract method')

    def kill(self, proc):
        raise NotImplementedError('abstract method')

    def check_finish(self, proc):
        raise NotImplementedError('abstract method')

    @classmethod
    def machine_arginfo(cls) -> Argument:
        """Generate the machine arginfo.

        Returns
        -------
        Argument
            machine arginfo
        """
        return Argument(
            cls.__name__, dict, sub_fields=cls.machine_subfields(),
            alias=[
                cls.__name__.lower(),
                cls.__name__.replace("Context", ""),
                cls.__name__.lower().replace("context", "")
            ])

    @classmethod
    def machine_subfields(cls) -> List[Argument]:
        """Generate the machine subfields.
        
        Returns
        -------
        list[Argument]
            machine subfields
        """
        doc_remote_profile = "The information used to maintain the connection with remote machine. This field is empty for this context."
        return [Argument("remote_profile", dict, optional=True, doc=doc_remote_profile)]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from os import path
import setuptools, datetime

readme_file = path.join(path.dirname(path.abspath(__file__)), 'README.md')
NAME="dpdispatcher"
SHORT_CMD="dpdisp"

try:
    from m2r import parse_from_file
    readme = parse_from_file(readme_file)
except ImportError:
    with open(readme_file) as f:
        readme = f.read()

today = datetime.date.today().strftime("%b-%d-%Y")
with open(path.join(NAME, '_date.py'), 'w') as fp :
    fp.write('date = \'%s\'' % today)

install_requires=['paramiko', 'dargs>=0.2.6', 'requests', 'tqdm']

setuptools.setup(
    name=NAME,
    use_scm_version={'write_to': 'dpdispatcher/_version.py'},
    setup_requires=['setuptools_scm'],
    author="Deep Modeling",
    author_email="",
    description="Python Dispatcher",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="",
    python_requires="~=3.6",
    packages=['dpdispatcher', 'dpdispatcher/dpcloudserver'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
    keywords='deep potential generator active learning deepmd-kit',
    install_requires=install_requires,    
    extras_require={
        'docs': ['sphinx', 'recommonmark', 'sphinx_rtd_theme>=1.0.0rc1', 'numpydoc'],
        "cloudserver": ["oss2", "tqdm"],
        ":python_version<'3.7'": ["typing_extensions"],
    },
        entry_points={
          'console_scripts': [
              SHORT_CMD+'= dpdispatcher.dpdisp:main']
   }
)



import sys, os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..' )))

from dpdispatcher.local_context import LocalSession
from dpdispatcher.local_context import LocalContext
from dpdispatcher.lazy_local_context import LazyLocalContext

from dpdispatcher.submission import Submission, Job, Task, Resources
from dpdispatcher.batch import Batch
# from dpdispatcher.pbs import PBS
from dpdispatcher.shell import Shell

local_session = LocalSession({'work_path':'temp2'})
local_context = LocalContext(local_root='test_shell_dir/', work_profile=local_session)
# lazy_local_context = LazyLocalContext(local_root='/home/fengbo/10_dpdispatcher/dpdispatcher/tests/temp3/0_md', work_profile=None)
shell = Shell(context=local_context)
# pbs = PBS(context=lazy_local_context)

resources = Resources(number_node=1, cpu_per_node=4, gpu_per_node=1, queue_name="V100_8_32", group_size=4) 
submission = Submission(work_base='0_md', resources=resources,  forward_common_files=['graph.pb'], backward_common_files=['submission.json']) #,  batch=PBS)
task1 = Task(command='lmp_serial -i input.lammps', task_work_path='bct-1', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'], task_need_resources=1)
task2 = Task(command='lmp_serial -i input.lammps', task_work_path='bct-2', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'], task_need_resources=0.25)
task3 = Task(command='lmp_serial -i input.lammps', task_work_path='bct-3', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'], task_need_resources=0.25)
task4 = Task(command='lmp_serial -i input.lammps', task_work_path='bct-4', forward_files=['conf.lmp', 'input.lammps'], backward_files=['log.lammps'], task_need_resources=0.5)
submission.register_task_list([task1, task2, task3, task4, ])
submission.generate_jobs()
submission.bind_batch(batch=shell)
# for job in submission.belonging_jobs:
#     job.job_to_json()
# print('111', submission)
# submission2 = Submission.recover_jobs_from_json('./jr.json')
# print('222', submission2)
# print(submission==submission2)
submission.run_submission()

# submission1.dump_jobs_fo_json()
# submission2 = Submission.submission_from_json('jsons/submission.json')
# print(677, submission==submission2)
# print(submission1.belonging_jobs)
# print(local_context)



# /usr/bin/python
# -*- encoding=utf-8 -*-

import os
import sys
from dpdispatcher.utils import run_cmd_with_all_output

class HDFS(object):
    '''Fundamental class for HDFS basic manipulation
    '''

    @staticmethod
    def exists(uri):
        '''Check existence of hdfs uri
        Returns: True on exists
        Raises: RuntimeError
        '''
        cmd = 'hadoop fs -test -e {uri}'.format(uri=uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            elif ret == 1:
                return False
            else:
                raise RuntimeError('Cannot check existence of hdfs uri[{}] '
                                   'with cmd[{}]; ret[{}] stdout[{}] stderr[{}]'.format(
                                    uri, cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot check existence of hdfs uri[{}] '
                               'with cmd[{}]'.format(uri, cmd)) from e

    @staticmethod
    def remove(uri):
        '''Check existence of hdfs uri
        Returns: True on exists
        Raises: RuntimeError
        '''

        cmd = 'hadoop fs -rm -r {uri}'.format(uri=uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError('Cannot remove hdfs uri[{}] '
                                   'with cmd[{}]; ret[{}] output[{}] stderr[{}]'.format(
                                    uri, cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot remove hdfs uri[{}] '
                               'with cmd[{}]'.format(uri, cmd)) from e

    @staticmethod
    def mkdir(uri):
        '''Make new hdfs directory
        Returns: True on success
        Raises: RuntimeError
        '''
        cmd = 'hadoop fs -mkdir -p {uri}'.format(uri=uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError('Cannot mkdir of hdfs uri[{}] '
                                   'with cmd[{}]; ret[{}] output[{}] stderr[{}]'.format(
                                    uri, cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot mkdir of hdfs uri[{}] '
                               'with cmd[{}]'.format(uri, cmd)) from e



    @staticmethod
    def copy_from_local(local_path, to_uri):
        '''
        Returns: True on success
        Raises: on unexpected error
        '''
        # Make sure local_path is accessible
        if not os.path.exists(local_path) or not os.access(local_path, os.R_OK):
            raise RuntimeError('try to access local_path[{}] '
                               'but failed'.format(local_path))
        cmd = 'hadoop fs -copyFromLocal -f {local} {remote}'.format(local=local_path,
                                                                    remote=to_uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True, out
            else:
                raise RuntimeError('Cannot copy local[{}] to remote[{}] with cmd[{}]; '
                                   'ret[{}] output[{}] stderr[{}]'.format(local_path, to_uri,
                                                                          cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot copy local[{}] to remote[{}] with cmd[{}]'
                               .format(local_path, to_uri, cmd)) from e

    @staticmethod
    def copy_to_local(from_uri, local_path):
        remote = ''
        if isinstance(from_uri, string_types):
            remote = from_uri
        elif isinstance(from_uri, list) or \
                isinstance(from_uri, tuple):
            remote = ' '.join(from_uri)
        cmd = 'hadoop fs -copyToLocal {remote} {local}'.format(remote=remote, local=local_path)

        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError('Cannot copy remote[{}] to local[{}] with cmd[{}]; '
                                   'ret[{}] output[{}] stderr[{}]'.format(from_uri, local_path,
                                                                          cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot copy remote[{}] to local[{}] with cmd[{}]'
                               .format(from_uri, local_path, cmd)) from e



    @staticmethod
    def read_hdfs_file(uri):
        cmd = 'hadoop fs -text {uri}'.format(uri=uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return out
            else:
                raise RuntimeError('Cannot read text from uri[{}]'
                                   'cmd [{}] ret[{}] output[{}] stderr[{}]'.format(uri, cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot read text from uri[{}]'
                               'cmd [{}]'.format(uri, cmd)) from e

    @staticmethod
    def move(from_uri, to_uri):
        cmd = 'hadoop fs -mv {furi} {turi}'.format(furi=from_uri, turi=to_uri)
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError('Cannot move from_uri[{}] to '
                                   'to_uri[{}] with cmd[{}]; '
                                   'ret[{}] output[{}] stderr[{}]'.format(from_uri, to_uri,
                                                                          cmd, ret, out, err))
        except Exception as e:
            raise RuntimeError('Cannot move from_uri[{}] to '
                               'to_uri[{}] with cmd[{}]'.format(from_uri, to_uri, cmd)) from e



#%%
import os
import sys
import uuid
import unittest

from dpdispatcher.dpcloudserver import api
from dpdispatcher.dpcloudserver.zip_file import zip_files
# import api
# from zip_file import zip_files


#%%
class DPTest(unittest.TestCase):

    test_data = {
        'job_type': 'indicate',
        'log_file': 'mylog',
        'command': '( echo aa && lmp -i input.lammps && sleep 900 ) > dp.log 2>&1',
        'backward_files': [],
        'job_name': 'dpdispatcher_lammps_test',
        'machine': {
            'platform': 'ali',
            'resources': {
                'gpu_type': '1 * NVIDIA P100',
                'cpu_num': 4,
                'mem_limit': 28,
                'time_limit': '2:00:00',
                'image_name': 'yfb-deepmd-kit-1.2.4-cuda10'
            }
        },
        'job_resources': 'http://dpcloudserver.oss-cn-shenzhen.aliyuncs.com/dpcloudserver/indicate/a657ff49722839f1ee54edeb3e9b1beb0ee5cc0e/a657ff49722839f1ee54edeb3e9b1beb0ee5cc0e.zip'
    }

    username = ''
    password = ''

    ENDPOINT = 'http://oss-cn-shenzhen.aliyuncs.com'
    BUCKET_NAME = 'dpcloudserver'

    @classmethod
    def setUpClass(cls):
        print('execute', sys._getframe().f_code.co_name)
       
    @classmethod
    def tearDownClass(cls):
        print('execute', sys._getframe().f_code.co_name)

    def setUp(self):
        print('execute', sys._getframe().f_code.co_name)
        api.login(self.username, self.password)

    def test_commit_job(self):
        print('----------', sys._getframe().f_code.co_name)
        file_uuid = uuid.uuid1().hex
        oss_task_zip = os.path.join('%s/%s/%s.zip' % ('indicate', file_uuid, file_uuid))
        zip_path = "/home/felix/workplace/22_dpdispatcher/dpdispatcher-yfb/dpdispatcher/dpcloudserver/t.txt"
        zip_task_file = zip_path + '.zip'
        zip_files(zip_path, zip_task_file, [])
        api.upload(oss_task_zip, zip_task_file, self.ENDPOINT, self.BUCKET_NAME)
        job_id = api.job_create(
            self.test_data['job_type'],
            self.test_data['job_resources'],
            self.test_data
            )
        tasks = api.get_tasks(job_id)
        print(tasks)

    def test_get_tasks(self):
        print('----------', sys._getframe().f_code.co_name)
        jobs = api.get_jobs()
        for j in jobs:
            tasks = api.get_tasks(j['id'])
            print(tasks)

    # def test_download(self):
    #     print('----------', sys._getframe().f_code.co_name)
    #     oss_path = 'dpcloudserver/indicate/abe0febc92ce11eb990800163e094dc5/abe0febc92ce11eb990800163e094dc5.zip'
    #     api.download(oss_path, "out.zip", self.ENDPOINT, self.BUCKET_NAME)

if __name__ == '__main__':
    suite = unittest.TestSuite()

    suite.addTest(DPTest("test_commit_job"))
    # suite.addTest(DPTest("test_get_tasks"))
    # suite.addTest(DPTest("test_download"))

    runner = unittest.TextTestRunner()
    runner.run(suite)




import os,sys,json,glob,shutil
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import LocalSession
from .context import setUpModule

class TestLocalSession(unittest.TestCase):
    def test_work_path(self):
        cwd = os.getcwd()        
        wp = LocalSession({'work_path' : cwd})
        self.assertTrue(os.path.abspath(cwd), wp.get_work_root())


        
